import os
import re
import shutil

import pytest
from click.testing import CliRunner
from freezegun import freeze_time

from great_expectations.cli import cli
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.util import gen_directory_tree_str, get_context
from tests.cli.test_cli import yaml
from tests.cli.test_datasource_sqlite import _add_datasource_and_credentials_to_context
from tests.cli.test_init_pandas import _delete_and_recreate_dir
from tests.cli.utils import assert_no_logging_messages_or_tracebacks

try:
    from unittest import mock
except ImportError:
    from unittest import mock


@pytest.fixture
def titanic_sqlite_db_file(sa, tmp_path_factory):
    temp_dir = str(tmp_path_factory.mktemp("foo_path"))
    fixture_db_path = file_relative_path(__file__, "../test_sets/titanic.db")

    db_path = os.path.join(temp_dir, "titanic.db")
    shutil.copy(fixture_db_path, db_path)

    engine = sa.create_engine(f"sqlite:///{db_path}", pool_recycle=3600)
    assert engine.execute("select count(*) from titanic").fetchall()[0] == (1313,)
    return db_path


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
@freeze_time("09/26/2019 13:42:41")
def test_cli_init_on_new_project(
    mock_webbrowser, caplog, monkeypatch, tmp_path_factory, titanic_sqlite_db_file, sa
):
    project_dir = str(tmp_path_factory.mktemp("test_cli_init_diff"))
    ge_dir = os.path.join(project_dir, "great_expectations")

    database_path = os.path.join(project_dir, "titanic.db")
    shutil.copy(titanic_sqlite_db_file, database_path)
    engine = sa.create_engine(f"sqlite:///{database_path}", pool_recycle=3600)

    inspector = sa.inspect(engine)

    # get the default schema and table for testing
    schemas = inspector.get_schema_names()
    default_schema = schemas[0]

    tables = [
        table_name for table_name in inspector.get_table_names(schema=default_schema)
    ]
    default_table = tables[0]

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    result = runner.invoke(
        cli,
        ["--v3-api", "init"],
        input=f"\n\n2\n6\ntitanic\n{engine.url}\n\n\n1\n{default_schema}\n{default_table}\nwarning\n\n\n\n",
        catch_exceptions=False,
    )
    stdout = result.output
    assert len(stdout) < 6000, "CLI output is unreasonably long."

    assert "Always know what to expect from your data" in stdout
    assert "What data would you like Great Expectations to connect to" in stdout
    assert "Which database backend are you using" in stdout
    assert "Give your new Datasource a short name" in stdout
    assert "What is the url/connection string for the sqlalchemy connection" in stdout
    assert "Attempting to connect to your database." in stdout
    assert "Great Expectations connected to your database" in stdout
    assert (
        "You have selected a datasource that is a SQL database. How would you like to specify the data?"
        in stdout
    )
    assert "Name the new Expectation Suite [main.titanic.warning]" in stdout
    assert (
        "Great Expectations will choose a couple of columns and generate expectations about them"
        in stdout
    )
    assert "Generating example Expectation Suite..." in stdout
    assert "Building" in stdout
    assert "Data Docs" in stdout
    assert "Great Expectations is now set up" in stdout

    context = get_context(context_root_dir=ge_dir)
    assert len(context.list_datasources()) == 1
    assert context.list_datasources()[0]["class_name"] == "SqlAlchemyDatasource"
    assert context.list_datasources()[0]["name"] == "titanic"

    first_suite = context.list_expectation_suites()[0]
    suite = context.get_expectation_suite(first_suite.expectation_suite_name)
    assert len(suite.expectations) == 14

    assert os.path.isdir(ge_dir)
    config_path = os.path.join(project_dir, "great_expectations/great_expectations.yml")
    assert os.path.isfile(config_path)

    config = yaml.load(open(config_path))
    data_source_class = config["datasources"]["titanic"]["data_asset_type"][
        "class_name"
    ]
    assert data_source_class == "SqlAlchemyDataset"

    obs_tree = gen_directory_tree_str(ge_dir)

    # Instead of monkey patching guids, just regex out the guids
    guid_safe_obs_tree = re.sub(
        r"[a-z0-9]{32}(?=\.(json|html))", "foobarbazguid", obs_tree
    )
    # print(guid_safe_obs_tree)
    assert (
        guid_safe_obs_tree
        == """\
great_expectations/
    .gitignore
    great_expectations.yml
    checkpoints/
    expectations/
        .ge_store_backend_id
        warning.json
    plugins/
        custom_data_docs/
            renderers/
            styles/
                data_docs_custom_styles.css
            views/
    profilers/
    uncommitted/
        config_variables.yml
        data_docs/
            local_site/
                index.html
                expectations/
                    warning.html
                static/
                    fonts/
                        HKGrotesk/
                            HKGrotesk-Bold.otf
                            HKGrotesk-BoldItalic.otf
                            HKGrotesk-Italic.otf
                            HKGrotesk-Light.otf
                            HKGrotesk-LightItalic.otf
                            HKGrotesk-Medium.otf
                            HKGrotesk-MediumItalic.otf
                            HKGrotesk-Regular.otf
                            HKGrotesk-SemiBold.otf
                            HKGrotesk-SemiBoldItalic.otf
                    images/
                        favicon.ico
                        glossary_scroller.gif
                        iterative-dev-loop.png
                        logo-long-vector.svg
                        logo-long.png
                        short-logo-vector.svg
                        short-logo.png
                        validation_failed_unexpected_values.gif
                    styles/
                        data_docs_custom_styles_template.css
                        data_docs_default_styles.css
                validations/
                    warning/
                        20190926T134241.000000Z/
                            20190926T134241.000000Z/
                                foobarbazguid.html
        validations/
            .ge_store_backend_id
            warning/
                20190926T134241.000000Z/
                    20190926T134241.000000Z/
                        foobarbazguid.json
"""
    )

    assert_no_logging_messages_or_tracebacks(caplog, result)

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        "{}/great_expectations/uncommitted/data_docs/local_site/validations/warning/".format(
            project_dir
        )
        in mock_webbrowser.call_args[0][0]
    )


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_cli_init_on_new_project_extra_whitespace_in_url(
    mock_webbrowser, caplog, monkeypatch, tmp_path_factory, titanic_sqlite_db_file, sa
):
    project_dir = str(tmp_path_factory.mktemp("test_cli_init_diff"))
    ge_dir = os.path.join(project_dir, "great_expectations")

    database_path = os.path.join(project_dir, "titanic.db")
    shutil.copy(titanic_sqlite_db_file, database_path)
    engine = sa.create_engine(f"sqlite:///{database_path}", pool_recycle=3600)
    engine_url_with_added_whitespace = "    " + str(engine.url) + "  "

    inspector = sa.inspect(engine)

    # get the default schema and table for testing
    schemas = inspector.get_schema_names()
    default_schema = schemas[0]

    tables = [
        table_name for table_name in inspector.get_table_names(schema=default_schema)
    ]
    default_table = tables[0]

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    result = runner.invoke(
        cli,
        ["--v3-api", "init"],
        input=f"\n\n2\n6\ntitanic\n{engine_url_with_added_whitespace}\n\n\n1\n{default_schema}\n{default_table}\nwarning\n\n\n\n",
        catch_exceptions=False,
    )
    stdout = result.output
    assert len(stdout) < 6000, "CLI output is unreasonably long."

    assert "Always know what to expect from your data" in stdout
    assert "What data would you like Great Expectations to connect to" in stdout
    assert "Which database backend are you using" in stdout
    assert "Give your new Datasource a short name" in stdout
    assert "What is the url/connection string for the sqlalchemy connection" in stdout
    assert "Attempting to connect to your database." in stdout
    assert "Great Expectations connected to your database" in stdout
    assert (
        "You have selected a datasource that is a SQL database. How would you like to specify the data?"
        in stdout
    )
    assert "Name the new Expectation Suite [main.titanic.warning]" in stdout
    assert (
        "Great Expectations will choose a couple of columns and generate expectations about them"
        in stdout
    )
    assert "Generating example Expectation Suite..." in stdout
    assert "Building" in stdout
    assert "Data Docs" in stdout
    assert "Great Expectations is now set up" in stdout

    context = get_context(context_root_dir=ge_dir)
    assert len(context.list_datasources()) == 1
    assert context.list_datasources() == [
        {
            "class_name": "SqlAlchemyDatasource",
            "name": "titanic",
            "module_name": "great_expectations.datasource",
            "credentials": {"url": str(engine.url)},
            "data_asset_type": {
                "class_name": "SqlAlchemyDataset",
                "module_name": "great_expectations.dataset",
            },
        }
    ]

    first_suite = context.list_expectation_suites()[0]
    suite = context.get_expectation_suite(first_suite.expectation_suite_name)
    assert len(suite.expectations) == 14

    assert os.path.isdir(ge_dir)
    config_path = os.path.join(project_dir, "great_expectations/great_expectations.yml")
    assert os.path.isfile(config_path)

    config = yaml.load(open(config_path))
    data_source_class = config["datasources"]["titanic"]["data_asset_type"][
        "class_name"
    ]
    assert data_source_class == "SqlAlchemyDataset"

    assert_no_logging_messages_or_tracebacks(caplog, result)

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        "{}/great_expectations/uncommitted/data_docs/local_site/validations/warning/".format(
            project_dir
        )
        in mock_webbrowser.call_args[0][0]
    )


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_init_on_existing_project_with_no_datasources_should_continue_init_flow_and_add_one(
    mock_webbrowser,
    caplog,
    monkeypatch,
    initialized_sqlite_project,
    titanic_sqlite_db_file,
    sa,
):
    project_dir = initialized_sqlite_project
    ge_dir = os.path.join(project_dir, FileDataContext.GX_DIR)

    _remove_all_datasources(ge_dir)
    os.remove(os.path.join(ge_dir, "expectations", "warning.json"))
    context = get_context(context_root_dir=ge_dir)
    assert not context.list_expectation_suites()

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    url = f"sqlite:///{titanic_sqlite_db_file}"

    inspector = sa.inspect(sa.create_engine(url))

    # get the default schema and table for testing
    schemas = inspector.get_schema_names()
    default_schema = schemas[0]

    tables = [
        table_name for table_name in inspector.get_table_names(schema=default_schema)
    ]
    default_table = tables[0]

    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input=f"\n\n2\n6\nsqlite\n{url}\n\n\n1\n{default_schema}\n{default_table}\nmy_suite\n\n\n\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        "{}/great_expectations/uncommitted/data_docs/local_site/validations/my_suite/".format(
            project_dir
        )
        in mock_webbrowser.call_args[0][0]
    )

    assert "Error: invalid input" not in stdout
    assert "Always know what to expect from your data" in stdout
    assert "What data would you like Great Expectations to connect to" in stdout
    assert (
        "Next, we will configure database credentials and store them in the `sqlite` section"
        in stdout
    )
    assert "What is the url/connection string for the sqlalchemy connection?" in stdout
    assert (
        "You have selected a datasource that is a SQL database. How would you like to specify the data?"
        in stdout
    )
    assert "Great Expectations connected to your database" in stdout
    assert "This looks like an existing project that" not in stdout

    config = _load_config_file(os.path.join(ge_dir, FileDataContext.GX_YML))
    assert "sqlite" in config["datasources"].keys()

    context = get_context(context_root_dir=ge_dir)
    assert context.list_datasources() == [
        {
            "class_name": "SqlAlchemyDatasource",
            "name": "sqlite",
            "module_name": "great_expectations.datasource",
            "credentials": {"url": url},
            "data_asset_type": {
                "class_name": "SqlAlchemyDataset",
                "module_name": "great_expectations.dataset",
            },
        }
    ]
    assert context.list_expectation_suites()[0].expectation_suite_name == "my_suite"
    assert len(context.list_expectation_suites()) == 1

    assert_no_logging_messages_or_tracebacks(caplog, result)


def _remove_all_datasources(ge_dir):
    config_path = os.path.join(ge_dir, FileDataContext.GX_YML)

    config = _load_config_file(config_path)
    config["datasources"] = {}

    with open(config_path, "w") as f:
        yaml.dump(config, f)

    context = get_context(context_root_dir=ge_dir)
    assert context.list_datasources() == []


def _load_config_file(config_path):
    assert os.path.isfile(config_path), "Config file is missing. Check path"

    with open(config_path) as f:
        read = f.read()
        config = yaml.load(read)

    assert isinstance(config, dict)
    return config


@pytest.fixture
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def initialized_sqlite_project(
    mock_webbrowser, caplog, monkeypatch, tmp_path_factory, titanic_sqlite_db_file, sa
):
    """This is an initialized project through the CLI."""
    project_dir = str(tmp_path_factory.mktemp("my_rad_project"))

    engine = sa.create_engine(f"sqlite:///{titanic_sqlite_db_file}", pool_recycle=3600)

    inspector = sa.inspect(engine)

    # get the default schema and table for testing
    schemas = inspector.get_schema_names()
    default_schema = schemas[0]

    tables = [
        table_name for table_name in inspector.get_table_names(schema=default_schema)
    ]
    default_table = tables[0]

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    result = runner.invoke(
        cli,
        ["--v3-api", "init"],
        input=f"\n\n2\n6\ntitanic\n{engine.url}\n\n\n1\n{default_schema}\n{default_table}\nwarning\n\n\n\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        f"{project_dir}/great_expectations/uncommitted/data_docs/local_site/validations/warning/"
        in mock_webbrowser.call_args[0][0]
    )

    assert_no_logging_messages_or_tracebacks(caplog, result)

    context = get_context(
        context_root_dir=os.path.join(project_dir, FileDataContext.GX_DIR)
    )
    assert isinstance(context, FileDataContext)
    assert len(context.list_datasources()) == 1
    assert context.list_datasources() == [
        {
            "class_name": "SqlAlchemyDatasource",
            "name": "titanic",
            "module_name": "great_expectations.datasource",
            "credentials": {"url": str(engine.url)},
            "data_asset_type": {
                "class_name": "SqlAlchemyDataset",
                "module_name": "great_expectations.dataset",
            },
        }
    ]
    return project_dir


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_init_on_existing_project_with_multiple_datasources_exist_do_nothing(
    mock_webbrowser,
    caplog,
    monkeypatch,
    initialized_sqlite_project,
    titanic_sqlite_db,
    empty_sqlite_db,
):
    project_dir = initialized_sqlite_project
    ge_dir = os.path.join(project_dir, FileDataContext.GX_DIR)

    context = get_context(context_root_dir=ge_dir)
    datasource_name = "wow_a_datasource"
    context = _add_datasource_and_credentials_to_context(
        context, datasource_name, empty_sqlite_db
    )
    assert len(context.list_datasources()) == 2

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="n\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 0

    assert "Error: invalid input" not in stdout

    assert "Always know what to expect from your data" in stdout
    assert "This looks like an existing project that" in stdout
    assert "appears complete" in stdout
    assert "Would you like to build & view this project's Data Docs" in stdout

    assert_no_logging_messages_or_tracebacks(caplog, result)


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_init_on_existing_project_with_datasource_with_existing_suite_offer_to_build_docs_answer_no(
    mock_webbrowser,
    caplog,
    monkeypatch,
    initialized_sqlite_project,
):
    project_dir = initialized_sqlite_project

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="n\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 0

    assert "Error: invalid input" not in stdout

    assert "Always know what to expect from your data" in stdout
    assert "This looks like an existing project that" in stdout
    assert "appears complete" in stdout
    assert "Would you like to build & view this project's Data Docs" in stdout

    assert_no_logging_messages_or_tracebacks(caplog, result)


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_init_on_existing_project_with_datasource_with_existing_suite_offer_to_build_docs_answer_yes(
    mock_webbrowser,
    caplog,
    monkeypatch,
    initialized_sqlite_project,
):
    project_dir = initialized_sqlite_project

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="\n\n",
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        "{}/great_expectations/uncommitted/data_docs/local_site/index.html".format(
            project_dir
        )
        in mock_webbrowser.call_args[0][0]
    )

    assert "Error: invalid input" not in stdout

    assert "Always know what to expect from your data" in stdout
    assert "This looks like an existing project that" in stdout
    assert "appears complete" in stdout
    assert "Would you like to build & view this project's Data Docs" in stdout

    assert_no_logging_messages_or_tracebacks(caplog, result)


@pytest.mark.xfail(
    reason="This command is not yet implemented for the modern API",
    run=True,
    strict=True,
)
@mock.patch("webbrowser.open", return_value=True, side_effect=None)
def test_init_on_existing_project_with_datasource_with_no_suite_create_one(
    mock_webbrowser, caplog, monkeypatch, initialized_sqlite_project, sa
):
    project_dir = initialized_sqlite_project
    ge_dir = os.path.join(project_dir, FileDataContext.GX_DIR)
    uncommitted_dir = os.path.join(ge_dir, "uncommitted")

    # mangle the setup to remove all traces of any suite
    expectations_dir = os.path.join(ge_dir, "expectations")
    data_docs_dir = os.path.join(uncommitted_dir, "data_docs")
    validations_dir = os.path.join(uncommitted_dir, "validations")

    _delete_and_recreate_dir(expectations_dir)
    _delete_and_recreate_dir(data_docs_dir)
    _delete_and_recreate_dir(validations_dir)

    context = get_context(context_root_dir=ge_dir)

    # get the datasource from data context
    all_datasources = context.list_datasources()
    datasource = all_datasources[0] if all_datasources else None

    # create a sqlalchemy engine using the URL of existing datasource
    engine = sa.create_engine(datasource.get("credentials", {}).get("url"))
    inspector = sa.inspect(engine)

    # get the default schema and table for testing
    schemas = inspector.get_schema_names()
    default_schema = schemas[0]

    tables = [
        table_name for table_name in inspector.get_table_names(schema=default_schema)
    ]
    default_table = tables[0]

    assert context.list_expectation_suites() == []

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(project_dir)
    with pytest.warns(
        UserWarning, match="Warning. An existing `great_expectations.yml` was found"
    ):
        result = runner.invoke(
            cli,
            ["--v3-api", "init"],
            input="\n1\n{schema}\n{table}\nsink_me\n\n\n\n".format(
                os.path.join(project_dir, "data/Titanic.csv"),
                schema=default_schema,
                table=default_table,
            ),
            catch_exceptions=False,
        )
    stdout = result.stdout

    assert result.exit_code == 0
    assert mock_webbrowser.call_count == 1
    assert (
        "{}/great_expectations/uncommitted/data_docs/local_site/validations/sink_me/".format(
            project_dir
        )
        in mock_webbrowser.call_args[0][0]
    )

    assert "Always know what to expect from your data" in stdout
    assert (
        "You have selected a datasource that is a SQL database. How would you like to specify the data?"
        in stdout
    )
    assert "Generating example Expectation Suite..." in stdout
    assert "The following Data Docs sites will be built" in stdout
    assert "Great Expectations is now set up" in stdout

    assert "Error: invalid input" not in stdout
    assert "This looks like an existing project that" not in stdout

    assert_no_logging_messages_or_tracebacks(caplog, result)

    context = get_context(context_root_dir=ge_dir)
    assert len(context.list_expectation_suites()) == 1



import os

from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest

sfAccount = os.environ.get("SNOWFLAKE_ACCOUNT")
sfUser = os.environ.get("SNOWFLAKE_USER")
sfPswd = os.environ.get("SNOWFLAKE_PW")
sfDatabase = os.environ.get("SNOWFLAKE_DATABASE")
sfSchema = os.environ.get("SNOWFLAKE_SCHEMA")
sfWarehouse = os.environ.get("SNOWFLAKE_WAREHOUSE")

CONNECTION_STRING = f"snowflake://{sfUser}:{sfPswd}@{sfAccount}/{sfDatabase}/{sfSchema}?warehouse={sfWarehouse}&application=great_expectations_oss"

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/snowflake_python_example.py datasource_config">
datasource_config = {
    "name": "my_snowflake_datasource",
    "class_name": "Datasource",
    "execution_engine": {
        "class_name": "SqlAlchemyExecutionEngine",
        "connection_string": "snowflake://<USER_NAME>:<PASSWORD>@<ACCOUNT_NAME>/<DATABASE_NAME>/<SCHEMA_NAME>?warehouse=<WAREHOUSE_NAME>&role=<ROLE_NAME>&application=great_expectations_oss",
    },
    "data_connectors": {
        "default_runtime_data_connector_name": {
            "class_name": "RuntimeDataConnector",
            "batch_identifiers": ["default_identifier_name"],
        },
        "default_inferred_data_connector_name": {
            "class_name": "InferredAssetSqlDataConnector",
            "include_schema_name": True,
        },
    },
}
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_config["execution_engine"]["connection_string"] = CONNECTION_STRING

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/snowflake_python_example.py test_yaml_config">
context.test_yaml_config(yaml.dump(datasource_config))
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/snowflake_python_example.py add_datasource">
context.add_datasource(**datasource_config)
# </snippet>

# First test for RuntimeBatchRequest using a query
batch_request = RuntimeBatchRequest(
    datasource_name="my_snowflake_datasource",
    data_connector_name="default_runtime_data_connector_name",
    data_asset_name="default_name",  # this can be anything that identifies this data
    runtime_parameters={
        "query": f"SELECT * from {sfSchema.lower()}.taxi_data LIMIT 10"
    },
    batch_identifiers={"default_identifier_name": "default_identifier"},
)

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)

# Second test for BatchRequest naming a table
# <snippet name="tests/integration/docusaurus/connecting_to_your_data/database/snowflake_python_example.py python batch_request name table">
batch_request = BatchRequest(
    datasource_name="my_snowflake_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name=f"{sfSchema.lower()}.taxi_data",  # this is the name of the table you want to retrieve
)
context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())
# </snippet>

# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_snowflake_datasource"]
assert f"{sfSchema.lower()}.taxi_data" in set(
    context.get_available_data_asset_names()["my_snowflake_datasource"][
        "default_inferred_data_connector_name"
    ]
)
validator.execution_engine.close()



import math
from datetime import datetime, timedelta
from timeit import timeit
from typing import Any, Dict

import dateutil
import pandas as pd
import pytest

from great_expectations.core import (
    ExpectationConfiguration,
    ExpectationValidationResult,
)
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.core.evaluation_parameters import (
    _deduplicate_evaluation_parameter_dependencies,
    find_evaluation_parameter_dependencies,
    parse_evaluation_parameter,
)
from great_expectations.data_context import DataContext
from great_expectations.exceptions import EvaluationParameterError


@pytest.mark.unit
def test_parse_evaluation_parameter():
    # Substitution alone is ok
    assert parse_evaluation_parameter("a", {"a": 1}) == 1
    assert (
        parse_evaluation_parameter(
            "urn:great_expectations:validations:blarg",
            {"urn:great_expectations:validations:blarg": 1},
        )
        == 1
    )

    # Very basic arithmetic is allowed as-is:
    assert parse_evaluation_parameter("1 + 1", {}) == 2

    # So is simple variable substitution:
    assert parse_evaluation_parameter("a + 1", {"a": 2}) == 3

    # URN syntax works
    assert (
        parse_evaluation_parameter(
            "urn:great_expectations:validations:source_patient_data.default"
            ":expect_table_row_count_to_equal.result.observed_value * 0.9",
            {
                "urn:great_expectations:validations:source_patient_data.default"
                ":expect_table_row_count_to_equal.result.observed_value": 10
            },
        )
        == 9
    )

    # bare decimal is accepted
    assert (
        parse_evaluation_parameter(
            "urn:great_expectations:validations:source_patient_data.default"
            ":expect_table_row_count_to_equal.result.observed_value * .9",
            {
                "urn:great_expectations:validations:source_patient_data.default"
                ":expect_table_row_count_to_equal.result.observed_value": 10
            },
        )
        == 9
    )

    # We have basic operations (trunc)
    assert (
        parse_evaluation_parameter(
            "urn:great_expectations:validations:source_patient_data.default"
            ":expect_table_row_count_to_equal.result.observed_value * 0.9",
            {
                "urn:great_expectations:validations:source_patient_data.default"
                ":expect_table_row_count_to_equal.result.observed_value": 11
            },
        )
        != 9
    )

    assert (
        parse_evaluation_parameter(
            "trunc(urn:great_expectations:validations:source_patient_data.default"
            ":expect_table_row_count_to_equal.result.observed_value * 0.9)",
            {
                "urn:great_expectations:validations:source_patient_data.default"
                ":expect_table_row_count_to_equal.result.observed_value": 11
            },
        )
        == 9
    )

    # Non GX URN syntax fails
    with pytest.raises(EvaluationParameterError) as err:
        parse_evaluation_parameter("urn:ieee:not_ge * 10", {"urn:ieee:not_ge": 1})
    assert "Parse Failure" in str(err.value)

    # Valid variables but invalid expression is no good
    with pytest.raises(EvaluationParameterError) as err:
        parse_evaluation_parameter("1 / a", {"a": 0})
    assert (
        "Error while evaluating evaluation parameter expression: division by zero"
        in str(err.value)
    )

    # It is okay to *substitute* strings in the expression...
    assert parse_evaluation_parameter("foo", {"foo": "bar"}) == "bar"

    # ...and to have whitespace in substituted values...
    assert parse_evaluation_parameter("foo", {"foo": "bar "}) == "bar "

    # ...but whitespace is *not* preserved from the parameter name if we evaluate it
    assert parse_evaluation_parameter("foo ", {"foo": "bar"}) == "bar"  # NOT "bar "

    # We can use multiple parameters...
    assert parse_evaluation_parameter("foo * bar", {"foo": 2, "bar": 3}) == 6

    # ...but we cannot leave *partially* evaluated expressions (phew!)
    with pytest.raises(EvaluationParameterError) as e:
        parse_evaluation_parameter("foo + bar", {"foo": 2})
    assert (
        "Error while evaluating evaluation parameter expression: Unknown string format: bar"
        in str(e.value)
    )


@pytest.mark.integration
def test_query_store_results_in_evaluation_parameters(data_context_with_query_store):
    TITANIC_ROW_COUNT = 1313  # taken from the titanic db conftest
    DISTINCT_TITANIC_ROW_COUNT = 4

    # parse_evaluation_parameters correctly resolves a stores URN
    res1 = parse_evaluation_parameter(
        parameter_expression="urn:great_expectations:stores:my_query_store:col_count",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert res1 == TITANIC_ROW_COUNT

    # and can handle an operator
    res2 = parse_evaluation_parameter(
        parameter_expression="urn:great_expectations:stores:my_query_store:col_count * 2",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert res2 == TITANIC_ROW_COUNT * 2

    # can even handle multiple operators
    res3 = parse_evaluation_parameter(
        parameter_expression="urn:great_expectations:stores:my_query_store:col_count * 0 + 100",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert res3 == 100

    # allows stores URNs with functions
    res4 = parse_evaluation_parameter(
        parameter_expression="cos(urn:great_expectations:stores:my_query_store:col_count)",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert math.isclose(math.cos(TITANIC_ROW_COUNT), res4)

    # multiple stores URNs can be used
    res5 = parse_evaluation_parameter(
        parameter_expression="urn:great_expectations:stores:my_query_store:col_count - urn:great_expectations:stores:my_query_store:dist_col_count",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert res5 == TITANIC_ROW_COUNT - DISTINCT_TITANIC_ROW_COUNT

    # complex expressions can combine operators, urns, and functions
    res6 = parse_evaluation_parameter(
        parameter_expression="abs(-urn:great_expectations:stores:my_query_store:col_count - urn:great_expectations:stores:my_query_store:dist_col_count)",
        evaluation_parameters=None,
        data_context=data_context_with_query_store,
    )
    assert res6 == TITANIC_ROW_COUNT + DISTINCT_TITANIC_ROW_COUNT


@pytest.mark.unit
def test_parser_timing():
    """We currently reuse the parser, clearing the stack between calls, which is about 10 times faster than not
    doing so. But these operations are really quick, so this may not be necessary."""
    assert (
        timeit(
            "parse_evaluation_parameter('x', {'x': 1})",
            setup="from great_expectations.core.evaluation_parameters import parse_evaluation_parameter",
            number=100,
        )
        < 1
    )


@pytest.mark.unit
def test_math_evaluation_paramaters():
    assert parse_evaluation_parameter("sin(2*PI)") == math.sin(math.pi * 2)
    assert parse_evaluation_parameter("cos(2*PI)") == math.cos(math.pi * 2)
    assert parse_evaluation_parameter("tan(2*PI)") == math.tan(math.pi * 2)


@pytest.mark.unit
def test_temporal_evaluation_parameters():
    # allow 1 second for "now" tolerance
    now = datetime.now()
    assert (
        (now - timedelta(weeks=1, seconds=3))
        < dateutil.parser.parse(
            parse_evaluation_parameter("now() - timedelta(weeks=1, seconds=2)")
        )
        < now - timedelta(weeks=1, seconds=1)
    )


@pytest.mark.unit
def test_temporal_evaluation_parameters_complex():
    # allow 1 second for "now" tolerance
    now = datetime.now()
    # Choosing "2*3" == 6 weeks shows we can parse an expression inside a kwarg.
    assert (
        (now - timedelta(weeks=2 * 3, seconds=3))
        < dateutil.parser.parse(
            parse_evaluation_parameter("now() - timedelta(weeks=2*3, seconds=2)")
        )
        < now - timedelta(weeks=2 * 3, seconds=1)
    )


@pytest.mark.unit
def test_find_evaluation_parameter_dependencies():
    parameter_expression = "(-3 * urn:great_expectations:validations:profile:expect_column_stdev_to_be_between.result.observed_value:column=norm) + urn:great_expectations:validations:profile:expect_column_mean_to_be_between.result.observed_value:column=norm"
    dependencies = find_evaluation_parameter_dependencies(parameter_expression)
    assert dependencies == {
        "urns": {
            "urn:great_expectations:validations:profile:expect_column_stdev_to_be_between.result.observed_value:column=norm",
            "urn:great_expectations:validations:profile:expect_column_mean_to_be_between.result.observed_value:column=norm",
        },
        "other": set(),
    }

    parameter_expression = "upstream_value * urn:great_expectations:validations:profile:expect_column_stdev_to_be_between.result.observed_value:column=norm"
    dependencies = find_evaluation_parameter_dependencies(parameter_expression)
    assert dependencies == {
        "urns": {
            "urn:great_expectations:validations:profile:expect_column_stdev_to_be_between.result.observed_value:column=norm",
        },
        "other": {"upstream_value"},
    }

    parameter_expression = "upstream_value"
    dependencies = find_evaluation_parameter_dependencies(parameter_expression)
    assert dependencies == {"urns": set(), "other": {"upstream_value"}}

    parameter_expression = "3 * upstream_value"
    dependencies = find_evaluation_parameter_dependencies(parameter_expression)
    assert dependencies == {"urns": set(), "other": {"upstream_value"}}

    parameter_expression = "3"
    dependencies = find_evaluation_parameter_dependencies(parameter_expression)
    assert dependencies == {"urns": set(), "other": set()}


@pytest.mark.unit
def test_deduplicate_evaluation_parameter_dependencies():
    dependencies = {
        "profile": [
            {
                "metric_kwargs_id": {
                    "column=norm": [
                        "expect_column_mean_to_be_between.result.observed_value"
                    ]
                }
            },
            {
                "metric_kwargs_id": {
                    "column=norm": [
                        "expect_column_stdev_to_be_between.result.observed_value"
                    ]
                }
            },
        ]
    }

    deduplicated = _deduplicate_evaluation_parameter_dependencies(dependencies)

    # For test, use set to ignore order
    deduplicated["profile"][0]["metric_kwargs_id"]["column=norm"] = set(
        deduplicated["profile"][0]["metric_kwargs_id"]["column=norm"]
    )

    assert {
        "profile": [
            {
                "metric_kwargs_id": {
                    "column=norm": {
                        "expect_column_mean_to_be_between.result.observed_value",
                        "expect_column_stdev_to_be_between.result.observed_value",
                    }
                }
            }
        ]
    } == deduplicated


@pytest.mark.integration
@pytest.mark.parametrize(
    "dataframe,evaluation_parameters,expectation_type,expectation_kwargs,expected_expectation_validation_result",
    [
        (
            pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
            (
                ("my_min", 1),
                ("my_max", 5),
            ),
            "expect_table_row_count_to_be_between",
            {
                "min_value": {
                    "$PARAMETER": "my_min",
                },
                "max_value": {
                    "$PARAMETER": "my_max",
                },
            },
            ExpectationValidationResult(
                expectation_config=ExpectationConfiguration(
                    expectation_type="expect_table_row_count_to_be_between",
                    kwargs={
                        "min_value": 1,
                        "max_value": 5,
                    },
                    meta={"substituted_parameters": {"min_value": 1, "max_value": 5}},
                    ge_cloud_id=None,
                ),
                meta={},
                exception_info={
                    "raised_exception": False,
                    "exception_traceback": None,
                    "exception_message": None,
                },
                success=True,
                result={"observed_value": 3},
            ),
        ),
        (
            pd.DataFrame(
                {
                    "my_date": [
                        datetime(year=2017, month=1, day=1),
                        datetime(year=2018, month=1, day=1),
                        datetime(year=2019, month=1, day=1),
                        datetime(year=2020, month=1, day=1),
                    ]
                }
            ),
            (
                ("my_min_date", datetime(2016, 12, 10)),
                ("my_max_date", datetime(2022, 12, 13)),
            ),
            "expect_column_values_to_be_between",
            {
                "column": "my_date",
                "min_value": {"$PARAMETER": "my_min_date"},
                "max_value": {"$PARAMETER": "my_max_date - timedelta(weeks=1)"},
            },
            ExpectationValidationResult(
                expectation_config=ExpectationConfiguration(
                    expectation_type="expect_column_values_to_be_between",
                    kwargs={
                        "column": "my_date",
                        "min_value": "2016-12-10T00:00:00",
                        "max_value": "2022-12-06T00:00:00",
                        "batch_id": "15fe04adb6ff20b9fc6eda486b7a36b7",
                    },
                    meta={
                        "substituted_parameters": {
                            "min_value": "2016-12-10T00:00:00",
                            "max_value": "2022-12-06T00:00:00",
                        }
                    },
                    ge_cloud_id=None,
                ),
                meta={},
                exception_info={
                    "raised_exception": False,
                    "exception_traceback": None,
                    "exception_message": None,
                },
                success=True,
                result={
                    "element_count": 4,
                    "unexpected_count": 0,
                    "unexpected_percent": 0.0,
                    "partial_unexpected_list": [],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 0.0,
                    "unexpected_percent_nonmissing": 0.0,
                },
            ),
        ),
    ],
)
def test_evaluation_parameters_for_between_expectations_parse_correctly(
    titanic_pandas_data_context_with_v013_datasource_with_checkpoints_v1_with_empty_store_stats_enabled: DataContext,
    dataframe: pd.DataFrame,
    evaluation_parameters: Dict[str, Any],
    expectation_type: str,
    expectation_kwargs: Dict[str, dict],
    expected_expectation_validation_result: ExpectationValidationResult,
):
    context = titanic_pandas_data_context_with_v013_datasource_with_checkpoints_v1_with_empty_store_stats_enabled

    expectation_suite_name = "test_suite"
    context.add_expectation_suite(expectation_suite_name=expectation_suite_name)

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="my_runtime_data_connector",
        data_asset_name="foo",
        runtime_parameters={"batch_data": dataframe},
        batch_identifiers={
            "pipeline_stage_name": "kickoff",
            "airflow_run_id": "1234",
        },
    )

    validator = context.get_validator(
        batch_request=batch_request,
        expectation_suite_name=expectation_suite_name,
    )

    for evaluation_parameter in evaluation_parameters:
        validator.set_evaluation_parameter(*evaluation_parameter)

    actual_expectation_validation_result = getattr(validator, expectation_type)(
        **expectation_kwargs
    )

    assert (
        actual_expectation_validation_result == expected_expectation_validation_result
    )


@pytest.mark.unit
def test_now_evaluation_parameter():
    """
    now() is unique in the fact that it is the only evaluation param built-in that has zero arity (takes no arguments).
    The following tests ensure that it is properly parsed and evaluated in a variety of contexts.
    """
    # By itself
    res = parse_evaluation_parameter("now()")
    assert dateutil.parser.parse(
        res
    ), "Provided evaluation parameter is not dateutil-parseable"

    # In conjunction with timedelta
    res = parse_evaluation_parameter("now() - timedelta(weeks=1)")
    assert dateutil.parser.parse(
        res
    ), "Provided evaluation parameter is not dateutil-parseable"

    # Require parens to actually invoke
    with pytest.raises(EvaluationParameterError):
        parse_evaluation_parameter("now")



import os
from collections import OrderedDict
from unittest import mock

import nbformat
import pytest
from click.testing import CliRunner
from nbconvert.preprocessors import ExecutePreprocessor

from great_expectations.cli import cli
from great_expectations.util import get_context
from tests.cli.utils import assert_no_logging_messages_or_tracebacks, escape_ansi


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
def test_cli_datasource_list(
    mock_emit, empty_data_context_stats_enabled, empty_sqlite_db, caplog, monkeypatch
):
    """Test an empty project and after adding a single datasource."""
    context = empty_data_context_stats_enabled

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    result = runner.invoke(
        cli,
        "--v3-api datasource list",
        catch_exceptions=False,
    )

    stdout = result.stdout.strip()
    assert "No Datasources found" in stdout
    assert context.list_datasources() == []

    datasource_name = "wow_a_datasource"
    _add_datasource_and_credentials_to_context(
        context, datasource_name, empty_sqlite_db
    )

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    result = runner.invoke(
        cli,
        ["--v3-api", "datasource", "list"],
        catch_exceptions=False,
    )
    expected_output = """\
Using v3 (Batch Request) API
1 Datasource found:

 - name: wow_a_datasource
   class_name: SqlAlchemyDatasource
""".strip()
    stdout = escape_ansi(result.stdout).strip()

    assert stdout == expected_output

    assert_no_logging_messages_or_tracebacks(caplog, result)
    anonymized_name: str = mock_emit.call_args_list[3][0][0]["event_payload"][
        "anonymized_name"
    ]

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.datasource.list.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.datasource.list.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "parent_class": "SqlAlchemyDatasource",
                },
                "event": "data_context.add_datasource",
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "datasource.sqlalchemy.connect",
                "event_payload": {
                    "anonymized_name": anonymized_name,
                    "sqlalchemy_dialect": "sqlite",
                },
                "success": True,
            }
        ),
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.datasource.list.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.datasource.list.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_count == len(expected_call_args_list)
    assert mock_emit.call_args_list == expected_call_args_list


def _add_datasource_and_credentials_to_context(context, datasource_name, sqlite_engine):
    original_datasources = context.list_datasources()

    url = str(sqlite_engine.url)
    credentials = {"url": url}
    context.save_config_variable(datasource_name, credentials)
    context.add_datasource(
        datasource_name,
        initialize=False,
        module_name="great_expectations.datasource",
        class_name="SqlAlchemyDatasource",
        data_asset_type={"class_name": "SqlAlchemyDataset"},
        credentials="${" + datasource_name + "}",
        batch_kwargs_generators={
            "default": {"class_name": "TableBatchKwargsGenerator"}
        },
    )

    expected_datasources = original_datasources
    expected_datasources.append(
        {
            "name": datasource_name,
            "class_name": "SqlAlchemyDatasource",
            "module_name": "great_expectations.datasource",
            "credentials": OrderedDict([("url", url)]),
            "data_asset_type": {"class_name": "SqlAlchemyDataset", "module_name": None},
            "batch_kwargs_generators": {
                "default": {"class_name": "TableBatchKwargsGenerator"}
            },
        }
    )

    assert context.list_datasources() == expected_datasources
    return context


def _add_datasource__with_two_generators_and_credentials_to_context(
    context, datasource_name, sqlite_engine
):
    original_datasources = context.list_datasources()

    url = str(sqlite_engine.url)
    credentials = {"url": url}
    context.save_config_variable(datasource_name, credentials)
    context.add_datasource(
        datasource_name,
        initialize=False,
        module_name="great_expectations.datasource",
        class_name="SqlAlchemyDatasource",
        data_asset_type={"class_name": "SqlAlchemyDataset"},
        credentials="${" + datasource_name + "}",
        batch_kwargs_generators={
            "default": {"class_name": "TableBatchKwargsGenerator"},
            "second_generator": {
                "class_name": "ManualBatchKwargsGenerator",
                "assets": {
                    "asset_one": [
                        {"partition_id": 1, "query": "select * from main.titanic"}
                    ]
                },
            },
        },
    )

    expected_datasources = original_datasources
    expected_datasources.append(
        {
            "name": datasource_name,
            "class_name": "SqlAlchemyDatasource",
            "module_name": "great_expectations.datasource",
            "credentials": {"url": url},
            "data_asset_type": {"class_name": "SqlAlchemyDataset", "module_name": None},
            "batch_kwargs_generators": {
                "default": {"class_name": "TableBatchKwargsGenerator"},
                "second_generator": {
                    "assets": {
                        "asset_one": [
                            {
                                "partition_id": 1,
                                "query": "select " "* " "from " "main.titanic",
                            }
                        ]
                    },
                    "class_name": "ManualBatchKwargsGenerator",
                },
            },
        }
    )

    assert context.list_datasources() == expected_datasources
    return context


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@mock.patch("subprocess.call", return_value=True, side_effect=None)
@pytest.mark.slow  # 6.81s
def test_cli_datasource_new_connection_string(
    mock_subprocess,
    mock_emit,
    empty_data_context_stats_enabled,
    empty_sqlite_db,
    caplog,
    monkeypatch,
):
    root_dir = empty_data_context_stats_enabled.root_directory
    context = empty_data_context_stats_enabled
    assert context.list_datasources() == []

    runner = CliRunner(mix_stderr=False)
    monkeypatch.chdir(os.path.dirname(context.root_directory))
    result = runner.invoke(
        cli,
        "--v3-api datasource new",
        input="2\n7\n",
        catch_exceptions=False,
    )
    stdout = result.stdout

    assert "What data would you like Great Expectations to connect to?" in stdout

    assert result.exit_code == 0

    uncommitted_dir = os.path.join(root_dir, context.GX_UNCOMMITTED_DIR)
    expected_notebook = os.path.join(uncommitted_dir, "datasource_new.ipynb")

    assert os.path.isfile(expected_notebook)
    mock_subprocess.assert_called_once_with(["jupyter", "notebook", expected_notebook])

    expected_call_args_list = [
        mock.call(
            {"event_payload": {}, "event": "data_context.__init__", "success": True}
        ),
        mock.call(
            {
                "event": "cli.datasource.new.begin",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.new_ds_choice",
                "event_payload": {
                    "type": "sqlalchemy",
                    "db": "other",
                    "api_version": "v3",
                },
                "success": True,
            }
        ),
        mock.call(
            {
                "event": "cli.datasource.new.end",
                "event_payload": {"api_version": "v3"},
                "success": True,
            }
        ),
    ]

    assert mock_emit.call_args_list == expected_call_args_list
    assert mock_emit.call_count == len(expected_call_args_list)

    # Run notebook
    with open(expected_notebook) as f:
        nb = nbformat.read(f, as_version=4)

    # Mock the user adding a connection string into the notebook by overwriting the right cell
    credentials_cell = nb["cells"][5]["source"]

    credentials = ("connection_string", "schema_name", "table_name")
    for credential in credentials:
        assert credential in credentials_cell

    # Replace placeholder with actual value to allow remainder of notebook to execute successfully
    nb["cells"][5]["source"] = credentials_cell.replace(
        "YOUR_CONNECTION_STRING", "sqlite://"
    )

    ep = ExecutePreprocessor(timeout=60, kernel_name="python3")
    ep.preprocess(nb, {"metadata": {"path": uncommitted_dir}})

    del context
    context = get_context(context_root_dir=root_dir)

    assert context.list_datasources() == [
        {
            "module_name": "great_expectations.datasource",
            "execution_engine": {
                "module_name": "great_expectations.execution_engine",
                "connection_string": "sqlite://",
                "class_name": "SqlAlchemyExecutionEngine",
            },
            "class_name": "Datasource",
            "data_connectors": {
                "default_runtime_data_connector_name": {
                    "batch_identifiers": ["default_identifier_name"],
                    "class_name": "RuntimeDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                },
                "default_inferred_data_connector_name": {
                    "class_name": "InferredAssetSqlDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                    "include_schema_name": True,
                    "introspection_directives": {
                        "schema_name": "YOUR_SCHEMA",
                    },
                },
                "default_configured_data_connector_name": {
                    "assets": {
                        "YOUR_TABLE_NAME": {
                            "class_name": "Asset",
                            "module_name": "great_expectations.datasource.data_connector.asset",
                            "schema_name": "YOUR_SCHEMA",
                        },
                    },
                    "class_name": "ConfiguredAssetSqlDataConnector",
                    "module_name": "great_expectations.datasource.data_connector",
                },
            },
            "name": "my_datasource",
        }
    ]

    assert_no_logging_messages_or_tracebacks(caplog, result)



# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py imports">
from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import BatchRequest
from great_expectations.profile.user_configurable_profiler import (
    UserConfigurableProfiler,
)

context = gx.get_context()
# </snippet>

# This utility is not for general use. It is only to support testing.
from tests.test_utils import load_data_into_test_database

# The following load & config blocks up until the batch requests are only to support testing.
MY_CONNECTION_STRING = "mysql+pymysql://root@localhost/test_ci"

PG_CONNECTION_STRING = "postgresql+psycopg2://postgres:@localhost/test_ci"

load_data_into_test_database(
    table_name="mysql_taxi_data",
    csv_path="./data/yellow_tripdata_sample_2019-01.csv",
    connection_string=MY_CONNECTION_STRING,
)

load_data_into_test_database(
    table_name="postgres_taxi_data",
    csv_path="./data/yellow_tripdata_sample_2019-01.csv",
    connection_string=PG_CONNECTION_STRING,
)

pg_datasource_config = {
    "name": "my_postgresql_datasource",
    "class_name": "Datasource",
    "execution_engine": {
        "class_name": "SqlAlchemyExecutionEngine",
        "connection_string": f"{PG_CONNECTION_STRING}",
    },
    "data_connectors": {
        "default_inferred_data_connector_name": {
            "class_name": "InferredAssetSqlDataConnector",
            "include_schema_name": True,
        },
    },
}

mysql_datasource_config = {
    "name": "my_mysql_datasource",
    "class_name": "Datasource",
    "execution_engine": {
        "class_name": "SqlAlchemyExecutionEngine",
        "connection_string": f"{MY_CONNECTION_STRING}",
    },
    "data_connectors": {
        "default_inferred_data_connector_name": {
            "class_name": "InferredAssetSqlDataConnector",
            "include_schema_name": True,
        },
    },
}

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml.
pg_datasource_config["execution_engine"]["connection_string"] = PG_CONNECTION_STRING

context.test_yaml_config(yaml.dump(pg_datasource_config))

context.add_datasource(**pg_datasource_config)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml.
mysql_datasource_config["execution_engine"]["connection_string"] = MY_CONNECTION_STRING

context.test_yaml_config(yaml.dump(mysql_datasource_config))

context.add_datasource(**mysql_datasource_config)

# Tutorial content resumes here.
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py mysql_batch_request">
mysql_batch_request = BatchRequest(
    datasource_name="my_mysql_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="test_ci.mysql_taxi_data",
)
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py pg_batch_request">
pg_batch_request = BatchRequest(
    datasource_name="my_postgresql_datasource",
    data_connector_name="default_inferred_data_connector_name",
    data_asset_name="public.postgres_taxi_data",
)
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py get_validator">
validator = context.get_validator(batch_request=mysql_batch_request)
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py profiler">
profiler = UserConfigurableProfiler(
    profile_dataset=validator,
    excluded_expectations=[
        "expect_column_quantile_values_to_be_between",
        "expect_column_mean_to_be_between",
    ],
)
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py build_suite">
expectation_suite_name = "compare_two_tables"
suite = profiler.build_suite()
suite.expectation_suite_name = expectation_suite_name
context.add_expectation_suite(expectation_suite=suite)
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py checkpoint_config">
my_checkpoint_name = "comparison_checkpoint"

yaml_config = f"""
name: {my_checkpoint_name}
config_version: 1.0
class_name: SimpleCheckpoint
run_name_template: "%Y%m%d-%H%M%S-my-run-name-template"
expectation_suite_name: {expectation_suite_name}
"""

context.add_or_update_checkpoint(**yaml.load(yaml_config))
# </snippet>
# <snippet name="tests/integration/docusaurus/expectations/advanced/user_configurable_profiler_cross_table_comparison.py run_checkpoint">
results = context.run_checkpoint(
    checkpoint_name=my_checkpoint_name, batch_request=pg_batch_request
)
# </snippet>

# Note to users: code below this line is only for integration testing -- ignore!

assert results["success"] is True
statistics = results["run_results"][list(results["run_results"].keys())[0]][
    "validation_result"
]["statistics"]
assert statistics["evaluated_expectations"] != 0
assert statistics["evaluated_expectations"] == statistics["successful_expectations"]
assert statistics["unsuccessful_expectations"] == 0
assert statistics["success_percent"] == 100.0



from typing import List

import requests

from tests.integration.usage_statistics.test_usage_statistics_messages import (
    valid_usage_statistics_messages,
)


def get_usage_stats_example_events() -> List:

    msg: dict = valid_usage_statistics_messages

    # remove top-level items from the messages
    keys: List = [k for k in msg.keys()]
    msg_list: List = []
    for key in keys:
        for item in msg[key]:
            msg_list.append(item)

    return msg_list


def get_gx_version() -> str:

    git_raw = "https://raw.githubusercontent.com"
    git_repo = "great-expectations"
    git_file = "great_expectations/develop/great_expectations/deployment_version"
    path = f"{git_raw}/{git_repo}/{git_file}"

    return requests.get(path).text.replace(".", "").replace("\n", "")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

from great_expectations.validator.computed_metric import MetricValue  # noqa: TCH001
from great_expectations.validator.exception_info import ExceptionInfo  # noqa: TCH001
from great_expectations.validator.metric_configuration import MetricConfiguration
from great_expectations.validator.validation_graph import ValidationGraph

if TYPE_CHECKING:
    from great_expectations.execution_engine import ExecutionEngine

logger = logging.getLogger(__name__)
logging.captureWarnings(True)

try:
    import pandas as pd
except ImportError:
    pd = None

    logger.debug(
        "Unable to load pandas; install optional pandas dependency for support."
    )


class MetricsCalculator:
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        show_progress_bars: bool = False,
    ) -> None:
        """
        MetricsCalculator accepts and processes metrics calculation requests.

        Args:
            execution_engine: ExecutionEngine to perform metrics computation.
            show_progress_bars: Directive for whether or not to show progress bars.
        """
        self._execution_engine: ExecutionEngine = execution_engine
        self._show_progress_bars: bool = show_progress_bars

    @property
    def show_progress_bars(self) -> bool:
        return self._show_progress_bars

    @show_progress_bars.setter
    def show_progress_bars(self, enable: bool) -> None:
        self._show_progress_bars = enable

    def columns(self, domain_kwargs: Optional[Dict[str, Any]] = None) -> List[str]:
        """
        Convenience method to run "table.columns" metric.
        """
        if domain_kwargs is None:
            domain_kwargs = {
                "batch_id": self._execution_engine.batch_manager.active_batch_data_id,
            }

        columns: List[str] = self.get_metric(
            metric=MetricConfiguration(
                metric_name="table.columns",
                metric_domain_kwargs=domain_kwargs,
            )
        )

        return columns

    def head(
        self,
        n_rows: int = 5,
        domain_kwargs: Optional[Dict[str, Any]] = None,
        fetch_all: bool = False,
    ) -> pd.DataFrame:
        """
        Convenience method to run "table.head" metric.
        """
        if domain_kwargs is None:
            domain_kwargs = {}

        if domain_kwargs.get("batch_id") is None:
            domain_kwargs[
                "batch_id"
            ] = self._execution_engine.batch_manager.active_batch_id

        df: pd.DataFrame = self.get_metric(
            metric=MetricConfiguration(
                metric_name="table.head",
                metric_domain_kwargs=domain_kwargs,
                metric_value_kwargs={
                    "n_rows": n_rows,
                    "fetch_all": fetch_all,
                },
            )
        )

        return df.reset_index(drop=True, inplace=False)

    def get_metric(
        self,
        metric: MetricConfiguration,
    ) -> Any:
        """return the value of the requested metric."""
        return self.get_metrics(
            metrics={metric.metric_name: metric},
        )[metric.metric_name]

    def get_metrics(
        self,
        metrics: Dict[str, MetricConfiguration],
    ) -> Dict[str, Any]:
        """
        Args:
            metrics: Dictionary of desired metrics to be resolved; metric_name is key and MetricConfiguration is value.

        Returns:
            Return Dictionary with requested metrics resolved, with metric_name as key and computed metric as value.
        """
        resolved_metrics: Dict[
            Tuple[str, str, str], MetricValue
        ] = self.compute_metrics(
            metric_configurations=list(metrics.values()),
            runtime_configuration=None,
            min_graph_edges_pbar_enable=0,
        )
        return {
            metric_configuration.metric_name: resolved_metrics[metric_configuration.id]
            for metric_configuration in metrics.values()
        }

    def compute_metrics(
        self,
        metric_configurations: List[MetricConfiguration],
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,
        # Set to low number (e.g., 3) to suppress progress bar for small graphs.
    ) -> Dict[Tuple[str, str, str], MetricValue]:
        """
        Args:
            metric_configurations: List of desired MetricConfiguration objects to be resolved.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").
            min_graph_edges_pbar_enable: Minumum number of graph edges to warrant showing progress bars.

        Returns:
            Dictionary with requested metrics resolved, with unique metric ID as key and computed metric as value.
        """
        graph: ValidationGraph = self.build_metric_dependency_graph(
            metric_configurations=metric_configurations,
            runtime_configuration=runtime_configuration,
        )
        resolved_metrics: Dict[Tuple[str, str, str], MetricValue]
        aborted_metrics_info: Dict[
            Tuple[str, str, str],
            Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]],
        ]
        (
            resolved_metrics,
            aborted_metrics_info,
        ) = self.resolve_validation_graph_and_handle_aborted_metrics_info(
            graph=graph,
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=min_graph_edges_pbar_enable,
        )
        return resolved_metrics

    def build_metric_dependency_graph(
        self,
        metric_configurations: List[MetricConfiguration],
        runtime_configuration: Optional[dict] = None,
    ) -> ValidationGraph:
        """
        Obtain domain and value keys for metrics and proceeds to add these metrics to the validation graph
        until all metrics have been added.

        Args:
            metric_configurations: List of "MetricConfiguration" objects, for which to build combined "ValidationGraph".
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").

        Returns:
            Resulting "ValidationGraph" object.
        """
        graph: ValidationGraph = ValidationGraph(
            execution_engine=self._execution_engine
        )

        metric_configuration: MetricConfiguration
        for metric_configuration in metric_configurations:
            graph.build_metric_dependency_graph(
                metric_configuration=metric_configuration,
                runtime_configuration=runtime_configuration,
            )

        return graph

    def resolve_validation_graph_and_handle_aborted_metrics_info(
        self,
        graph: ValidationGraph,
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,
        # Set to low number (e.g., 3) to suppress progress bar for small graphs.
    ) -> Tuple[
        Dict[Tuple[str, str, str], MetricValue],
        Dict[
            Tuple[str, str, str],
            Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]],
        ],
    ]:
        """
        Args:
            graph: "ValidationGraph" object, containing "metric_edge" structures with "MetricConfiguration" objects.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").
            min_graph_edges_pbar_enable: Minumum number of graph edges to warrant showing progress bars.

        Returns:
            Dictionary with requested metrics resolved, with unique metric ID as key and computed metric as value.
        """
        resolved_metrics: Dict[Tuple[str, str, str], MetricValue]
        aborted_metrics_info: Dict[
            Tuple[str, str, str],
            Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]],
        ]
        (resolved_metrics, aborted_metrics_info,) = self.resolve_validation_graph(
            graph=graph,
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=min_graph_edges_pbar_enable,
        )

        if aborted_metrics_info:
            logger.warning(
                f"Exceptions\n{str(aborted_metrics_info)}\noccurred while resolving metrics."
            )

        return (
            resolved_metrics,
            aborted_metrics_info,
        )

    def resolve_validation_graph(
        self,
        graph: ValidationGraph,
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,
        # Set to low number (e.g., 3) to suppress progress bar for small graphs.
    ) -> Tuple[
        Dict[Tuple[str, str, str], MetricValue],
        Dict[
            Tuple[str, str, str],
            Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]],
        ],
    ]:
        """
        Calls "ValidationGraph.resolve()" method with supplied arguments.

        Args:
            graph: "ValidationGraph" object, containing "metric_edge" structures with "MetricConfiguration" objects.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").
            min_graph_edges_pbar_enable: Minumum number of graph edges to warrant showing progress bars.

        Returns:
            Dictionary with requested metrics resolved, with unique metric ID as key and computed metric as value.
            Aborted metrics information, with metric ID as key.
        """
        resolved_metrics: Dict[Tuple[str, str, str], MetricValue]
        aborted_metrics_info: Dict[
            Tuple[str, str, str],
            Dict[str, Union[MetricConfiguration, Set[ExceptionInfo], int]],
        ]
        resolved_metrics, aborted_metrics_info = graph.resolve(
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=min_graph_edges_pbar_enable,
            show_progress_bars=self._show_progress_bars,
        )
        return resolved_metrics, aborted_metrics_info



# <snippet name="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py imports">
import boto3
import yaml
from awsglue.context import GlueContext
from pyspark.context import SparkContext

import great_expectations as gx
from great_expectations.checkpoint import SimpleCheckpoint
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.data_context.types.base import (
    DataContextConfig,
    S3StoreBackendDefaults,
)
from great_expectations.util import get_context

# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py set up gx">
sc = SparkContext.getOrCreate()
glueContext = GlueContext(sc)
spark = glueContext.spark_session
s3_client = boto3.client("s3")
response = s3_client.get_object(
    Bucket="bucket", Key="bucket/great_expectations/great_expectations.yml"
)
config_file = yaml.safe_load(response["Body"])
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py connect to data">
config = DataContextConfig(
    config_version=config_file["config_version"],
    datasources=config_file["datasources"],
    expectations_store_name=config_file["expectations_store_name"],
    validations_store_name=config_file["validations_store_name"],
    evaluation_parameter_store_name=config_file["evaluation_parameter_store_name"],
    plugins_directory="/great_expectations/plugins",
    validation_operators=config_file["validation_operators"],
    stores=config_file["stores"],
    data_docs_sites=config_file["data_docs_sites"],
    config_variables_file_path=config_file["config_variables_file_path"],
    anonymous_usage_statistics=config_file["anonymous_usage_statistics"],
    checkpoint_store_name=config_file["checkpoint_store_name"],
    store_backend_defaults=S3StoreBackendDefaults(
        default_bucket_name=config_file["data_docs_sites"]["s3_site"]["store_backend"][
            "bucket"
        ]
    ),
)
context_gx = get_context(project_config=config)
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py create expectations">
expectation_suite_name = "suite_name"
suite = context_gx.add_expectation_suite(expectation_suite_name)
batch_request = RuntimeBatchRequest(
    datasource_name="spark_s3",
    data_asset_name="datafile_name",
    batch_identifiers={"runtime_batch_identifier_name": "default_identifier"},
    data_connector_name="default_inferred_data_connector_name",
    runtime_parameters={"path": "s3a://bucket_name/path_to_file.format"},
)
validator = context_gx.get_validator(
    batch_request=batch_request,
    expectation_suite_name=expectation_suite_name,
)
print(validator.head())
validator.expect_column_values_to_not_be_null(
    column="passenger_count"
)  ## add some test
validator.save_expectation_suite(discard_failed_expectations=False)
# </snippet>

# <snippet name="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py validate your data">
checkpoint_config = {
    "class_name": "SimpleCheckpoint",
    "validations": [
        {
            "batch_request": batch_request,
            "expectation_suite_name": expectation_suite_name,
        }
    ],
}

checkpoint = SimpleCheckpoint(
    f"_tmp_checkpoint_{expectation_suite_name}", context_gx, **checkpoint_config
)
results = checkpoint.run(result_format="SUMMARY", run_name="test")
validation_result_identifier = results.list_validation_result_identifiers()[0]
# </snippet>



#!/usr/bin/env python3

import logging
import socket
import sys
import time
import uuid

import pandas as pd

from great_expectations.data_context.types.base import DataContextConfig
from great_expectations.util import get_context

logging.basicConfig(level=logging.INFO)


def guard(*args, **kwargs):
    raise ConnectionError("Internet Access is Blocked!")


def main(
    data_context_id=None,
    nap_duration=1,
    block_network=False,
    enable_usage_statistics=True,
):
    if data_context_id is None:
        data_context_id = str(uuid.uuid4())

    if block_network:
        socket.socket = guard

    print("Beginning to construct a DataContext.")
    config = DataContextConfig(
        config_version=2,
        datasources={"pandas": {"class_name": "PandasDatasource"}},
        expectations_store_name="expectations",
        validations_store_name="validations",
        evaluation_parameter_store_name="evaluation_parameters",
        plugins_directory=None,
        validation_operators={
            "action_list_operator": {
                "class_name": "ActionListValidationOperator",
                "action_list": [
                    {
                        "name": "store_validation_result",
                        "action": {"class_name": "StoreValidationResultAction"},
                    }
                ],
            }
        },
        stores={
            "expectations": {"class_name": "ExpectationsStore"},
            "validations": {"class_name": "ValidationsStore"},
            "evaluation_parameters": {"class_name": "EvaluationParameterStore"},
        },
        data_docs_sites={},
        config_variables_file_path=None,
        anonymous_usage_statistics={
            "enabled": enable_usage_statistics,
            # Leaving data_context_id as none would cause a new id to be generated
            "data_context_id": data_context_id,
            # This will be overridden when tests set an environment variable
            "usage_statistics_url": "https://qa.stats.greatexpectations.io/great_expectations/v1/usage_statistics",
        },
        commented_map=None,
    )
    context = get_context(config)
    print("Done constructing a DataContext.")
    print("Building a suite and validating.")
    df = pd.DataFrame({"a": [1, 2, 3]})
    context.add_expectation_suite("testing.batch")
    batch = context.get_batch(
        batch_kwargs={"datasource": "pandas", "dataset": df},
        expectation_suite_name="testing.batch",
    )
    batch.expect_column_values_to_be_between("a", 0, 5)
    batch.expect_column_to_exist("a")
    batch.save_expectation_suite()
    res = context.run_validation_operator("action_list_operator", [batch])
    print(res)
    print("Beginning a nap.")
    time.sleep(nap_duration)
    print("Ending a long nap.")


if __name__ == "__main__":
    data_context_id = sys.argv[1]

    try:
        nap_duration = int(sys.argv[2])
    except IndexError:
        nap_duration = 1
    except ValueError:
        print("Unrecognized sleep duration. Using 1 second.")
        nap_duration = 1

    try:
        res = sys.argv[3]
        if res in ["y", "yes", "True", "true", "t", "T"]:
            block_network = True
        else:
            block_network = False
    except IndexError:
        block_network = False
    except ValueError:
        print("Unrecognized value for block_network. Setting to false.")
        block_network = False

    try:
        res = sys.argv[4]
        if res in ["y", "yes", "True", "true", "t", "T"]:
            enable_usage_statistics = True
        else:
            enable_usage_statistics = False
    except IndexError:
        enable_usage_statistics = True
    except ValueError:
        print("Unrecognized value for usage_statistics_enabled. Setting to True.")
        enable_usage_statistics = True

    ge_logger = logging.getLogger("great_expectations")
    ge_logger.setLevel(logging.DEBUG)
    main(
        data_context_id,
        nap_duration,
        block_network=block_network,
        enable_usage_statistics=enable_usage_statistics,
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from typing_extensions import NotRequired, TypedDict

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.data_context_key import DataContextKey
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.exceptions import ClassInstantiationError, DataContextError

if TYPE_CHECKING:
    from great_expectations.core.configuration import AbstractConfig

logger = logging.getLogger(__name__)


class StoreConfigTypedDict(TypedDict):
    # NOTE: TypeDict values may be incomplete, update as needed
    class_name: str
    module_name: NotRequired[str]
    store_backend: dict


class Store:
    """A store is responsible for reading and writing Great Expectations objects
    to appropriate backends. It provides a generic API that the DataContext can
    use independently of any particular ORM and backend.

    An implementation of a store will generally need to define the following:
      - serialize
      - deserialize
      - _key_class (class of expected key type)

    All keys must have a to_tuple() method.
    """

    _key_class = DataContextKey

    def __init__(
        self,
        store_backend: Optional[dict] = None,
        runtime_environment: Optional[dict] = None,
        store_name: str = "no_store_name",
    ) -> None:
        """
        Runtime environment may be necessary to instantiate store backend elements.
        Args:
            store_backend:
            runtime_environment:
            store_name: store name given in the DataContextConfig (via either in-code or yaml configuration)
        """
        if store_backend is None:
            store_backend = {"class_name": "InMemoryStoreBackend"}
        self._store_name = store_name
        logger.debug("Building store_backend.")
        module_name = "great_expectations.data_context.store"
        self._store_backend = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment or {},
            config_defaults={
                "module_name": module_name,
                "store_name": self._store_name,
            },
        )
        if not self._store_backend:
            raise ClassInstantiationError(
                module_name=module_name, package_name=None, class_name=store_backend
            )
        if not isinstance(self._store_backend, StoreBackend):
            raise DataContextError(
                "Invalid StoreBackend configuration: expected a StoreBackend instance."
            )
        self._use_fixed_length_key = self._store_backend.fixed_length_key

    def ge_cloud_response_json_to_object_dict(self, response_json: Dict) -> Dict:
        """
        This method takes full json response from GX cloud and outputs a dict appropriate for
        deserialization into a GX object
        """
        return response_json

    def _validate_key(self, key: DataContextKey) -> None:
        # STORE_BACKEND_ID_KEY always validated
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return
        elif not isinstance(key, self.key_class):
            raise TypeError(
                f"key must be an instance of {self.key_class.__name__}, not {type(key)}"
            )

    @property
    def cloud_mode(self) -> bool:
        return isinstance(self._store_backend, GXCloudStoreBackend)

    @property
    def ge_cloud_mode(self) -> bool:
        # <GX_RENAME> Deprecated 0.15.37
        return self.cloud_mode

    @property
    def store_backend(self) -> StoreBackend:
        return self._store_backend

    @property
    def store_name(self) -> str:
        return self._store_name

    @property
    def store_backend_id(self) -> str:
        """
        Report the store_backend_id of the currently-configured StoreBackend
        Returns:
            store_backend_id which is a UUID(version=4)
        """
        return self._store_backend.store_backend_id

    @property
    def key_class(self) -> Type[DataContextKey]:
        if self.cloud_mode:
            return GXCloudIdentifier
        return self._key_class

    @property
    def store_backend_id_warnings_suppressed(self) -> str:
        """
        Report the store_backend_id of the currently-configured StoreBackend, suppressing warnings for invalid configurations.
        Returns:
            store_backend_id which is a UUID(version=4)
        """
        return self._store_backend.store_backend_id_warnings_suppressed

    @property
    def config(self) -> dict:
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def serialize(self, value: Any) -> Any:
        return value

    # noinspection PyMethodMayBeStatic
    def key_to_tuple(self, key: DataContextKey) -> Tuple[str, ...]:
        if self._use_fixed_length_key:
            return key.to_fixed_length_tuple()
        return key.to_tuple()

    def tuple_to_key(self, tuple_: Tuple[str, ...]) -> DataContextKey:
        if tuple_ == StoreBackend.STORE_BACKEND_ID_KEY:
            return StoreBackend.STORE_BACKEND_ID_KEY[0]  # type: ignore[return-value]
        if self._use_fixed_length_key:
            return self.key_class.from_fixed_length_tuple(tuple_)
        return self.key_class.from_tuple(tuple_)

    # noinspection PyMethodMayBeStatic
    def deserialize(self, value: Any) -> Any:
        return value

    def get(
        self, key: DataContextKey | GXCloudIdentifier | ConfigurationIdentifier
    ) -> Optional[Any]:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.get(key)

        if self.cloud_mode:
            self._validate_key(key)
            value = self._store_backend.get(self.key_to_tuple(key))
            # TODO [Robby] MER-285: Handle non-200 http errors
            if value:
                value = self.ge_cloud_response_json_to_object_dict(response_json=value)
        else:
            self._validate_key(key)
            value = self._store_backend.get(self.key_to_tuple(key))

        if value:
            return self.deserialize(value)

        return None

    def set(self, key: DataContextKey, value: Any, **kwargs) -> None:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.set(key, value, **kwargs)

        self._validate_key(key)
        return self._store_backend.set(
            self.key_to_tuple(key), self.serialize(value), **kwargs
        )

    def add(self, key: DataContextKey, value: Any, **kwargs) -> None:
        """
        Essentially `set` but validates that a given key-value pair does not already exist.
        """
        self._validate_key(key)
        return self._store_backend.add(
            self.key_to_tuple(key), self.serialize(value), **kwargs
        )

    def update(self, key: DataContextKey, value: Any, **kwargs) -> None:
        """
        Essentially `set` but validates that a given key-value pair does already exist.
        """
        self._validate_key(key)
        return self._store_backend.update(
            self.key_to_tuple(key), self.serialize(value), **kwargs
        )

    def add_or_update(self, key: DataContextKey, value: Any, **kwargs) -> None:
        """
        Conditionally calls `add` or `update` based on the presence of the given key.
        """
        self._validate_key(key)
        return self._store_backend.add_or_update(
            self.key_to_tuple(key), self.serialize(value), **kwargs
        )

    def list_keys(self) -> List[DataContextKey]:
        keys_without_store_backend_id = [
            key
            for key in self._store_backend.list_keys()
            if not key == StoreBackend.STORE_BACKEND_ID_KEY
        ]
        return [self.tuple_to_key(key) for key in keys_without_store_backend_id]

    def has_key(self, key: DataContextKey) -> bool:
        if key == StoreBackend.STORE_BACKEND_ID_KEY:
            return self._store_backend.has_key(key)  # noqa: W601
        else:
            if self._use_fixed_length_key:
                return self._store_backend.has_key(  # noqa: W601
                    key.to_fixed_length_tuple()
                )
            return self._store_backend.has_key(key.to_tuple())  # noqa: W601

    def self_check(self, pretty_print: bool) -> None:
        NotImplementedError(
            f"The test method is not implemented for Store class {self.__class__.__name__}."
        )

    def _build_key_from_config(self, config: AbstractConfig) -> DataContextKey:
        id: Optional[str] = None
        # Chetan - 20220831 - Explicit fork in logic to cover legacy behavior (particularly around Checkpoints).
        # Will be removed as part of the effort to rename `ge_cloud_id` to `id` across the codebase.
        if hasattr(config, "ge_cloud_id"):
            id = config.ge_cloud_id
        if hasattr(config, "id"):
            id = config.id

        name: Optional[str] = None
        if hasattr(config, "name"):
            name = config.name

        return self.store_backend.build_key(name=name, id=id)

    @staticmethod
    def build_store_from_config(
        store_name: Optional[str] = None,
        store_config: StoreConfigTypedDict | dict | None = None,
        module_name: str = "great_expectations.data_context.store",
        runtime_environment: Optional[dict] = None,
    ) -> Store:
        if store_config is None or module_name is None:
            raise gx_exceptions.StoreConfigurationError(
                "Cannot build a store without both a store_config and a module_name"
            )

        try:
            config_defaults: dict = {
                "store_name": store_name,
                "module_name": module_name,
            }
            new_store = instantiate_class_from_config(
                config=store_config,
                runtime_environment=runtime_environment,
                config_defaults=config_defaults,
            )
        except gx_exceptions.DataContextError as e:
            logger.critical(
                f"Error {e} occurred while attempting to instantiate a store."
            )
            class_name: str = store_config["class_name"]
            module_name = store_config.get("module_name", module_name)
            raise gx_exceptions.ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=class_name,
            ) from e

        return new_store



from typing import List

import jsonschema
import pytest

from great_expectations.core.usage_statistics.schemas import (
    anonymized_datasource_schema,
    anonymized_domain_builder_schema,
    anonymized_expectation_configuration_builder_schema,
    anonymized_init_payload_schema,
    anonymized_parameter_builder_schema,
    anonymized_rule_based_profiler_run_schema,
    anonymized_rule_schema,
    anonymized_string_schema,
    anonymized_usage_statistics_record_schema,
)


def test_anonymized_name_validation():
    string = "aa41efe0a1b3eeb9bf303e4561ff8392"
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_string_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(string)

    with pytest.raises(jsonschema.ValidationError):
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_string_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(string[:5])


def test_anonymized_datasource_validation():
    record = {
        "anonymized_name": "aa41efe0a1b3eeb9bf303e4561ff8392",
        "parent_class": "hello",
    }
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_datasource_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(record)

    record = {
        "anonymized_name": "aa41efe0a1b3eeb9bf303e4561ff8392",
        "parent_class": "hello",
        "anonymized_class": "aa41efe0a1b3eeb9bf303e4561ff8392",
    }
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_datasource_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(record)
    record = {
        "anonymized_name": "aa41efe0a1b3eeb9bf303e4561ff8392",
        "parent_class": "SparkDFDataset",
        "anonymized_class": "aa41efe0a1b3eeb9bf303e4561ff8392",
        "sqlalchemy_dialect": "postgres",
    }
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_datasource_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(record)


def test_init_payload_validation():
    payload = {
        "platform.system": "Darwin",
        "platform.release": "19.3.0",
        "version_info": "sys.version_info(major=3, minor=7, micro=4, releaselevel='final', serial=0)",
        "anonymized_datasources": [
            {
                "anonymized_name": "f57d8a6edae4f321b833384801847498",
                "parent_class": "SqlAlchemyDatasource",
                "sqlalchemy_dialect": "postgresql",
            }
        ],
        "anonymized_stores": [
            {
                "anonymized_name": "078eceafc1051edf98ae2f911484c7f7",
                "parent_class": "ExpectationsStore",
                "anonymized_store_backend": {
                    "parent_class": "TupleFilesystemStoreBackend"
                },
            },
            {
                "anonymized_name": "313cbd9858dd92f3fc2ef1c10ab9c7c8",
                "parent_class": "ValidationsStore",
                "anonymized_store_backend": {
                    "parent_class": "TupleFilesystemStoreBackend"
                },
            },
            {
                "anonymized_name": "2d487386aa7b39e00ed672739421473f",
                "parent_class": "EvaluationParameterStore",
                "anonymized_store_backend": {"parent_class": "InMemoryStoreBackend"},
            },
        ],
        "anonymized_validation_operators": [
            {
                "anonymized_name": "99d14cc00b69317551690fb8a61aca94",
                "parent_class": "ActionListValidationOperator",
                "anonymized_action_list": [
                    {
                        "anonymized_name": "5a170e5b77c092cc6c9f5cf2b639459a",
                        "parent_class": "StoreValidationResultAction",
                    },
                    {
                        "anonymized_name": "0fffe1906a8f2a5625a5659a848c25a3",
                        "parent_class": "StoreEvaluationParametersAction",
                    },
                    {
                        "anonymized_name": "101c746ab7597e22b94d6e5f10b75916",
                        "parent_class": "UpdateDataDocsAction",
                    },
                ],
            }
        ],
        "anonymized_data_docs_sites": [
            {
                "parent_class": "SiteBuilder",
                "anonymized_name": "eaf0cf17ad63abf1477f7c37ad192700",
                "anonymized_store_backend": {
                    "parent_class": "TupleFilesystemStoreBackend"
                },
                "anonymized_site_index_builder": {
                    "parent_class": "DefaultSiteIndexBuilder",
                },
            }
        ],
        "anonymized_expectation_suites": [
            {
                "anonymized_name": "238e99998c7674e4ff26a9c529d43da4",
                "expectation_count": 8,
                "anonymized_expectation_type_counts": {
                    "expect_column_value_lengths_to_be_between": 1,
                    "expect_table_row_count_to_be_between": 1,
                    "expect_column_values_to_not_be_null": 2,
                    "expect_column_distinct_values_to_be_in_set": 1,
                    "expect_column_kl_divergence_to_be_less_than": 1,
                    "expect_table_column_count_to_equal": 1,
                    "expect_table_columns_to_match_ordered_list": 1,
                },
            }
        ],
    }
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_init_payload_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(payload)


def test_run_val_op_message():
    message = {
        "event_payload": {
            "anonymized_operator_name": "50daa62a8739db21009f452f7e36153b",
        },
        "event": "data_context.run_validation_operator",
        "success": True,
        "version": "1.0.0",
        "event_time": "2020-03-26T23:02:17.932Z",
        "data_context_id": "705dd2a2-27f8-470f-9ebe-e7058fd7a534",
        "data_context_instance_id": "4f6deb55-8fbd-4131-9f97-b42b0902eae5",
        "ge_version": "0.9.7+203.ge3a97f44.dirty",
    }
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_usage_statistics_record_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(message)


def test_anonymized_domain_builder_schema():
    messages: List[dict] = [
        {
            "parent_class": "TableDomainBuilder",
        },
        {
            "parent_class": "TableDomainBuilder",
            "anonymized_class": "6a72f3728393d6519a197cffdccd50ff",
        },
        {
            "parent_class": "TableDomainBuilder",
            "anonymized_batch_request": {
                "anonymized_batch_request_required_top_level_properties": {
                    "anonymized_data_asset_name": "cdk128c5824b698c22b441ada61022d4",
                    "anonymized_data_connector_name": "456a3221fc4b65014d061cce4a71782e",
                    "anonymized_datasource_name": "eg78ebde1957385a02d8736cd2c9a6d9",
                }
            },
        },
    ]

    for message in messages:
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_domain_builder_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(message)


def test_anonymized_parameter_builder_schema() -> None:
    messages: List[dict] = [
        {
            "anonymized_name": "bdea123asdh13r90asdfh1209adsfh0a",
            "parent_class": "RegexPatternStringParameterBuilder",
        },
        {
            "anonymized_name": "bdea123asdh13r90asdfh1209adsfh0a",
            "parent_class": "RegexPatternStringParameterBuilder",
            "anonymized_class": "6a72f3728393d6519a197cffdccd50ff",
        },
        {
            "anonymized_name": "bdea123asdh13r90asdfh1209adsfh0a",
            "parent_class": "RegexPatternStringParameterBuilder",
            "anonymized_batch_request": {
                "anonymized_batch_request_required_top_level_properties": {
                    "anonymized_data_asset_name": "cdk128c5824b698c22b441ada61022d4",
                    "anonymized_data_connector_name": "456a3221fc4b65014d061cce4a71782e",
                    "anonymized_datasource_name": "eg78ebde1957385a02d8736cd2c9a6d9",
                }
            },
        },
    ]

    for message in messages:
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_parameter_builder_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(message)


def test_anonymized_expectation_configuration_builder_schema() -> None:
    messages: List[dict] = [
        {
            "parent_class": "DefaultExpectationConfigurationBuilder",
            "expectation_type": "expect_column_distinct_values_to_be_in_set",
        },
        {
            "parent_class": "DefaultExpectationConfigurationBuilder",
            "expectation_type": "expect_column_distinct_values_to_be_in_set",
            "anonymized_condition": "dsfj23931jasdf0923j2309asdfja0wd",
        },
        {
            "parent_class": "DefaultExpectationConfigurationBuilder",
            "anonymized_class": "6a72f3728393d6519a197cffdccd50ff",
            "expectation_type": "expect_column_distinct_values_to_be_in_set",
        },
        {
            "parent_class": "DefaultExpectationConfigurationBuilder",
            "anonymized_expectation_type": "asba9ad239asee92d092fjf2309f3j32",
        },
    ]

    for message in messages:
        jsonschema.validate(
            message, anonymized_expectation_configuration_builder_schema
        )


def test_anonymized_rule_schema():
    message = {
        "anonymized_name": "asdf932jwdf823r9ozsf9j20zsdfjas9",
        "anonymized_domain_builder": {
            "parent_class": "TableDomainBuilder",
            "anonymized_class": "6a72f3728393d6519a197cffdccd50ff",
        },
        "anonymized_parameter_builders": [
            {
                "anonymized_name": "bdea123asdh13r90asdfh1209adsfh0a",
                "parent_class": "RegexPatternStringParameterBuilder",
                "anonymized_batch_request": {
                    "anonymized_batch_request_required_top_level_properties": {
                        "anonymized_data_asset_name": "cdk128c5824b698c22b441ada61022d4",
                        "anonymized_data_connector_name": "456a3221fc4b65014d061cce4a71782e",
                        "anonymized_datasource_name": "eg78ebde1957385a02d8736cd2c9a6d9",
                    }
                },
            },
        ],
        "anonymized_expectation_configuration_builders": [
            {
                "parent_class": "DefaultExpectationConfigurationBuilder",
                "anonymized_expectation_type": "asba9ad239asee92d092fjf2309f3j32",
            },
        ],
    }

    jsonschema.validators.Draft202012Validator(
        schema=anonymized_rule_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(message)


def test_anonymized_rule_based_profiler_validation():
    message: dict = {
        "anonymized_name": "5b6c98e19e21e77191fb071bb9e80070",
        "anonymized_rules": [
            {
                "anonymized_domain_builder": {"parent_class": "TableDomainBuilder"},
                "anonymized_expectation_configuration_builders": [
                    {
                        "parent_class": "DefaultExpectationConfigurationBuilder",
                        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    }
                ],
                "anonymized_name": "5a83f3728393d6519a197cffdccd50ff",
                "anonymized_parameter_builders": [
                    {
                        "anonymized_name": "9349ed253aba01f4ecf190af61018a11",
                        "anonymized_class": "dasfj238fefasfa90sdf23j39202f2j2",
                        "parent_class": "MetricMultiBatchParameterBuilder",
                    }
                ],
            },
            {
                "anonymized_domain_builder": {"parent_class": "TableDomainBuilder"},
                "anonymized_expectation_configuration_builders": [
                    {
                        "parent_class": "DefaultExpectationConfigurationBuilder",
                        "anonymized_expectation_type": "239asdfjaew832hg20sdjwd9922e9e2u",
                    }
                ],
                "anonymized_name": "0bac2cecbb0cf8bb704e86710941434e",
                "anonymized_parameter_builders": [
                    {
                        "anonymized_name": "b7719efec76c6ebe30230fc1ec023beb",
                        "parent_class": "MetricMultiBatchParameterBuilder",
                    }
                ],
            },
        ],
        "config_version": 1.0,
        "rule_count": 2,
        "variable_count": 1,
    }

    # Populated payload (valid)
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_rule_based_profiler_run_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(message)

    # Remove mandatory field (invalid)
    with pytest.raises(jsonschema.ValidationError) as e:
        message.pop("variable_count")
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_rule_based_profiler_run_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(message)

    expected_error_msg = "'variable_count' is a required property"
    assert any(
        expected_error_msg in args
        for args in [context.args for context in e.value.context]
    )

    # Empty payload (valid)
    message.clear()
    jsonschema.validators.Draft202012Validator(
        schema=anonymized_rule_based_profiler_run_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    ).validate(message)


def test_uuid_format_validation_success():
    message = {
        "event_payload": {
            "anonymized_operator_name": "50daa62a8739db21009f452f7e36153b",
        },
        "event": "data_context.run_validation_operator",
        "success": True,
        "version": "1.0.0",
        "event_time": "2020-03-26T23:02:17.932Z",
        "data_context_id": "705dd2a2-27f8-470f-9ebe-e7058fd7a534",
        "data_context_instance_id": "4f6deb55-8fbd-4131-9f97-b42b0902eae5",
        "ge_version": "0.9.7+203.ge3a97f44.dirty",
    }
    try:
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_usage_statistics_record_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(message)
    except jsonschema.ValidationError as e:
        assert False, e.message


def test_uuid_format_validation_failure():
    message = {
        "event_payload": {
            "anonymized_operator_name": "50daa62a8739db21009f452f7e36153b",
        },
        "event": "data_context.run_validation_operator",
        "success": True,
        "version": "1.0.0",
        "event_time": "2020-03-26T23:02:17.932Z",
        "data_context_id": "705dd2a2-27f8-470f-9ebe-e7058fd7a534",
        "data_context_instance_id": "4f6deb55-8fbd-4131-9f97-b42b0902eae5 other_invalid_text",
        "ge_version": "0.9.7+203.ge3a97f44.dirty",
    }
    with pytest.raises(jsonschema.ValidationError) as e:
        jsonschema.validators.Draft202012Validator(
            schema=anonymized_usage_statistics_record_schema,
            format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
        ).validate(message)
    assert (
        "'4f6deb55-8fbd-4131-9f97-b42b0902eae5 other_invalid_text' is not a 'uuid'"
        in e.value.message
    )



from typing import Callable, Tuple
from unittest import mock

import pytest

from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.data_context.data_context import DataContext
from great_expectations.data_context.types.base import DataContextConfig, GXCloudConfig
from great_expectations.data_context.types.resource_identifiers import GXCloudIdentifier
from great_expectations.rule_based_profiler.config.base import (
    ruleBasedProfilerConfigSchema,
)
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler
from great_expectations.util import get_context
from tests.data_context.conftest import MockResponse


@pytest.fixture
def profiler_id() -> str:
    return "45f8a554-c90e-464e-8dff-f4008eb2cbb8"


@pytest.fixture
def profiler_without_id(profiler_rules: dict) -> RuleBasedProfiler:
    return RuleBasedProfiler(
        "my_profiler",
        config_version=1.0,
        rules=profiler_rules,
    )


@pytest.fixture
def profiler_with_id(profiler_id: str, profiler_rules: dict) -> RuleBasedProfiler:
    return RuleBasedProfiler(
        "my_profiler",
        config_version=1.0,
        id=profiler_id,
        rules=profiler_rules,
    )


@pytest.fixture
def mocked_get_response(
    mock_response_factory: Callable,
    profiler_id: str,
    profiler_with_id: RuleBasedProfiler,
) -> Callable[[], MockResponse]:
    profiler_config_dict = profiler_with_id.config.to_json_dict()
    created_by_id = "c06ac6a2-52e0-431e-b878-9df624edc8b8"
    organization_id = "046fe9bc-c85b-4e95-b1af-e4ce36ba5384"

    def _mocked_get_response(*args, **kwargs):
        return mock_response_factory(
            {
                "data": {
                    "id": profiler_id,
                    "attributes": {
                        "profiler": profiler_config_dict,
                        "created_at": "2022-08-02T17:55:45.107550",
                        "created_by_id": created_by_id,
                        "deleted": False,
                        "deleted_at": None,
                        "desc": None,
                        "organization_id": f"{organization_id}",
                        "updated_at": "2022-08-02T17:55:45.107550",
                    },
                    "links": {
                        "self": f"/organizations/{organization_id}/profilers/{profiler_id}"
                    },
                    "type": "profiler",
                },
            },
            200,
        )

    return _mocked_get_response


@pytest.fixture
def mocked_post_response(
    mock_response_factory: Callable, profiler_id: str
) -> Callable[[], MockResponse]:
    def _mocked_post_response(*args, **kwargs):
        return mock_response_factory(
            {
                "data": {
                    "id": profiler_id,
                }
            },
            201,
        )

    return _mocked_post_response


@pytest.mark.cloud
@pytest.mark.integration
def test_profiler_save_with_existing_profiler_retrieves_obj_with_id_from_store(
    empty_base_data_context_in_cloud_mode: CloudDataContext,
    profiler_with_id: RuleBasedProfiler,
    mocked_get_response: Callable[[], MockResponse],
    ge_cloud_base_url: str,
    ge_cloud_organization_id: str,
) -> None:
    """
    What does this test do and why?

    `DataContext.save_profiler()` should take in an input profiler with an id, save it to the GX Cloud backend,
    and return the same profiler (id and all).
    """
    context = empty_base_data_context_in_cloud_mode

    with mock.patch("requests.Session.put", autospec=True) as mock_put, mock.patch(
        "requests.Session.get", autospec=True, side_effect=mocked_get_response
    ) as mock_get, pytest.deprecated_call():
        return_profiler = context.save_profiler(profiler=profiler_with_id)

    profiler_id = profiler_with_id.ge_cloud_id
    expected_profiler_config = ruleBasedProfilerConfigSchema.dump(
        profiler_with_id.config
    )

    mock_put.assert_called_once_with(
        mock.ANY,  # requests.Session object
        f"{ge_cloud_base_url}/organizations/{ge_cloud_organization_id}/profilers/{profiler_id}",
        json={
            "data": {
                "attributes": {
                    "organization_id": ge_cloud_organization_id,
                    "profiler": expected_profiler_config,
                },
                "id": profiler_id,
                "type": "profiler",
            },
        },
    )

    mock_get.assert_called_once_with(
        mock.ANY,  # requests.Session object
        f"{ge_cloud_base_url}/organizations/{ge_cloud_organization_id}/profilers/{profiler_id}",
        params=None,
    )

    assert return_profiler.ge_cloud_id == profiler_with_id.ge_cloud_id


@pytest.mark.cloud
@pytest.mark.integration
def test_profiler_save_with_new_profiler_retrieves_obj_with_id_from_store(
    empty_base_data_context_in_cloud_mode: CloudDataContext,
    profiler_without_id: RuleBasedProfiler,
    profiler_id: str,
    mocked_get_response: Callable[[], MockResponse],
    mocked_post_response: Callable[[], MockResponse],
    ge_cloud_base_url: str,
    ge_cloud_organization_id: str,
) -> None:
    """
    What does this test do and why?

    `DataContext.save_profiler()` should take in an input profiler without an id, save it to the GX Cloud backend,
    and return the same profiler with the Cloud generated id.
    """
    context = empty_base_data_context_in_cloud_mode

    with mock.patch(
        "requests.Session.post", autospec=True, side_effect=mocked_post_response
    ) as mock_post, mock.patch(
        "requests.Session.get", autospec=True, side_effect=mocked_get_response
    ) as mock_get, pytest.deprecated_call():
        return_profiler = context.save_profiler(profiler=profiler_without_id)

    expected_profiler_config = ruleBasedProfilerConfigSchema.dump(
        profiler_without_id.config
    )

    mock_post.assert_called_once_with(
        mock.ANY,  # requests.Session object
        f"{ge_cloud_base_url}/organizations/{ge_cloud_organization_id}/profilers",
        json={
            "data": {
                "type": "profiler",
                "attributes": {
                    "organization_id": ge_cloud_organization_id,
                    "profiler": expected_profiler_config,
                },
            },
        },
    )

    mock_get.assert_called_once_with(
        mock.ANY,  # requests.Session object
        f"{ge_cloud_base_url}/organizations/{ge_cloud_organization_id}/profilers/{profiler_id}",
        params=None,
    )

    assert return_profiler.ge_cloud_id == profiler_id


@pytest.mark.e2e
@pytest.mark.cloud
@mock.patch("great_expectations.data_context.DataContext._save_project_config")
@pytest.mark.xfail(
    strict=False,
    reason="GX Cloud E2E tests are failing due to env vars not being consistently recognized by Docker; x-failing for purposes of 0.15.22 release",
)
def test_cloud_backed_data_context_add_profiler_e2e(
    mock_save_project_config: mock.MagicMock,
    profiler_rules: dict,
) -> None:
    context = DataContext(cloud_mode=True)

    name = "oss_test_profiler"
    config_version = 1.0
    rules = profiler_rules

    profiler = context.add_profiler(
        name=name, config_version=config_version, rules=rules
    )

    ge_cloud_id = profiler.ge_cloud_id

    profiler_stored_in_cloud = context.get_profiler(ge_cloud_id=ge_cloud_id)

    assert profiler.ge_cloud_id == profiler_stored_in_cloud.ge_cloud_id
    assert profiler.to_json_dict() == profiler_stored_in_cloud.to_json_dict()


@pytest.fixture
def profiler_names_and_ids() -> Tuple[Tuple[str, str], Tuple[str, str]]:
    profiler_name_1 = "Test Profiler 1"
    profiler_id_1 = "43996c1d-ac32-493d-8e68-eb41e37ac039"
    profiler_name_2 = "Test Profiler 2"
    profiler_id_2 = "94604e27-22d6-4264-89fe-a52c9e4b654d"

    profiler_1 = (profiler_name_1, profiler_id_1)
    profiler_2 = (profiler_name_2, profiler_id_2)
    return profiler_1, profiler_2


@pytest.fixture
def mock_get_all_profilers_json(
    profiler_names_and_ids: Tuple[Tuple[str, str], Tuple[str, str]]
) -> dict:
    profiler_1, profiler_2 = profiler_names_and_ids
    profiler_name_1, profiler_id_1 = profiler_1
    profiler_name_2, profiler_id_2 = profiler_2
    mock_json = {
        "data": [
            {
                "attributes": {
                    "created_by_id": "df665fc4-1891-4ef7-9a12-a0c46015c92c",
                    "organization_id": "841d7bd5-e321-4951-9b60-deb086203c7d",
                    "profiler": {
                        "name": profiler_name_1,
                        "rules": {
                            "my_rule_for_numeric_columns": {
                                "domain_builder": {
                                    "batch_request": "$variables.my_last_month_sales_batch_request",
                                    "class_name": "SemanticTypeColumnDomainBuilder",
                                    "semantic_types": ["numeric"],
                                },
                                "expectation_configuration_builders": [
                                    {
                                        "class_name": "DefaultExpectationConfigurationBuilder",
                                        "column": "$domain.domain_kwargs.column",
                                        "expectation_type": "expect_column_values_to_be_between",
                                        "max_value": "$parameter.my_column_max.value",
                                        "min_value": "$parameter.my_column_min.value",
                                        "mostly": "$variables.mostly_default",
                                    }
                                ],
                                "parameter_builders": [
                                    {
                                        "batch_request": "$variables.my_last_month_sales_batch_request",
                                        "class_name": "MetricParameterBuilder",
                                        "metric_domain_kwargs": "$domain.domain_kwargs",
                                        "metric_name": "column.min",
                                        "parameter_name": "my_column_min",
                                    },
                                ],
                            }
                        },
                        "variables": {},
                    },
                },
                "id": profiler_id_1,
                "type": "profiler",
            },
            {
                "attributes": {
                    "created_by_id": "df665fc4-1891-4ef7-9a12-a0c46015c92c",
                    "organization_id": "841d7bd5-e321-4951-9b60-deb086203c7d",
                    "profiler": {
                        "name": profiler_name_2,
                        "rules": {
                            "my_rule_for_numeric_columns": {
                                "domain_builder": {
                                    "batch_request": "$variables.my_last_month_sales_batch_request",
                                    "class_name": "SemanticTypeColumnDomainBuilder",
                                    "semantic_types": ["numeric"],
                                },
                                "expectation_configuration_builders": [
                                    {
                                        "class_name": "DefaultExpectationConfigurationBuilder",
                                        "column": "$domain.domain_kwargs.column",
                                        "expectation_type": "expect_column_values_to_be_between",
                                        "max_value": "$parameter.my_column_max.value",
                                        "min_value": "$parameter.my_column_min.value",
                                        "mostly": "$variables.mostly_default",
                                    }
                                ],
                                "parameter_builders": [
                                    {
                                        "batch_request": "$variables.my_last_month_sales_batch_request",
                                        "class_name": "MetricParameterBuilder",
                                        "metric_domain_kwargs": "$domain.domain_kwargs",
                                        "metric_name": "column.min",
                                        "parameter_name": "my_column_min",
                                    },
                                ],
                            }
                        },
                        "variables": {},
                    },
                },
                "id": profiler_id_2,
                "type": "profiler",
            },
        ]
    }
    return mock_json


@pytest.mark.unit
@pytest.mark.cloud
def test_list_profilers(
    empty_ge_cloud_data_context_config: DataContextConfig,
    ge_cloud_config: GXCloudConfig,
    profiler_names_and_ids: Tuple[Tuple[str, str], Tuple[str, str]],
    mock_get_all_profilers_json: dict,
) -> None:
    project_path_name = "foo/bar/baz"

    context = get_context(
        project_config=empty_ge_cloud_data_context_config,
        context_root_dir=project_path_name,
        cloud_base_url=ge_cloud_config.base_url,
        cloud_organization_id=ge_cloud_config.organization_id,
        cloud_access_token=ge_cloud_config.access_token,
        cloud_mode=True,
    )

    profiler_1, profiler_2 = profiler_names_and_ids
    profiler_name_1, profiler_id_1 = profiler_1
    profiler_name_2, profiler_id_2 = profiler_2

    with mock.patch("requests.Session.get", autospec=True) as mock_get:
        mock_get.return_value = mock.Mock(
            status_code=200, json=lambda: mock_get_all_profilers_json
        )
        profilers = context.list_profilers()

    assert profilers == [
        GXCloudIdentifier(
            resource_type=GXCloudRESTResource.PROFILER,
            cloud_id=profiler_id_1,
            resource_name=profiler_name_1,
        ),
        GXCloudIdentifier(
            resource_type=GXCloudRESTResource.PROFILER,
            cloud_id=profiler_id_2,
            resource_name=profiler_name_2,
        ),
    ]



import os
import random
from typing import Callable, List, Tuple
from unittest import mock

import pandas as pd
import pytest

from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context import BaseDataContext
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.data_context.data_context import DataContext
from great_expectations.data_context.types.base import DataContextConfig
from great_expectations.validator.validator import Validator

from great_expectations.data_context.store import (  # isort:skip
    ExpectationsStore,
    ValidationsStore,
    EvaluationParameterStore,
)

yaml = YAMLHandler()


@pytest.fixture()
def basic_in_memory_data_context_config_just_stores():
    return DataContextConfig(
        config_version=3.0,
        plugins_directory=None,
        evaluation_parameter_store_name="evaluation_parameter_store",
        expectations_store_name="expectations_store",
        datasources={},
        stores={
            "expectations_store": {"class_name": "ExpectationsStore"},
            "evaluation_parameter_store": {"class_name": "EvaluationParameterStore"},
            "validation_result_store": {"class_name": "ValidationsStore"},
        },
        validations_store_name="validation_result_store",
        data_docs_sites={},
        validation_operators={},
    )


@pytest.fixture()
def basic_in_memory_data_context_just_stores(
    basic_in_memory_data_context_config_just_stores,
):
    return BaseDataContext(
        project_config=basic_in_memory_data_context_config_just_stores
    )


def test_instantiation_and_basic_stores(
    basic_in_memory_data_context_just_stores,
    basic_in_memory_data_context_config_just_stores,
):
    context: BaseDataContext = basic_in_memory_data_context_just_stores
    # TODO <WILL> - config is basic_in_memory_data_context_config_just_stores + global overrides. Add test for this
    assert len(context.stores) == 3

    assert context.expectations_store_name == "expectations_store"
    assert isinstance(context.expectations_store, ExpectationsStore)

    assert context.validations_store_name == "validation_result_store"
    assert isinstance(context.validations_store, ValidationsStore)

    assert context.evaluation_parameter_store_name == "evaluation_parameter_store"
    assert isinstance(context.evaluation_parameter_store, EvaluationParameterStore)

    # TODO : validation_store
    # TODO : checkpoint store
    # TODO : metric store?


def test_config_variables(basic_in_memory_data_context_just_stores):
    # nothing instantiated yet
    assert basic_in_memory_data_context_just_stores.config_variables == {}


def test_list_stores(basic_in_memory_data_context_just_stores):
    assert basic_in_memory_data_context_just_stores.list_stores() == [
        {"class_name": "ExpectationsStore", "name": "expectations_store"},
        {
            "class_name": "EvaluationParameterStore",
            "name": "evaluation_parameter_store",
        },
        {"class_name": "ValidationsStore", "name": "validation_result_store"},
    ]


def test_add_store(basic_in_memory_data_context_just_stores):
    store_name: str = "my_new_expectations_store"
    store_config: dict = {"class_name": "ExpectationsStore"}
    basic_in_memory_data_context_just_stores.add_store(
        store_name=store_name, store_config=store_config
    )
    assert basic_in_memory_data_context_just_stores.list_stores() == [
        {"class_name": "ExpectationsStore", "name": "expectations_store"},
        {
            "class_name": "EvaluationParameterStore",
            "name": "evaluation_parameter_store",
        },
        {"class_name": "ValidationsStore", "name": "validation_result_store"},
        {"class_name": "ExpectationsStore", "name": "my_new_expectations_store"},
    ]


def test_list_active_stores(basic_in_memory_data_context_just_stores):
    """
    Active stores are identified by the following keys:
        expectations_store_name,
        validations_store_name,
        evaluation_parameter_store_name,
        checkpoint_store_name
        profiler_store_name
    Therefore the test also test that the list_active_stores() output doesn't change after a store named
    `my_new_expectations_store` is added
    """
    expected_store_list: List[dict] = [
        {"class_name": "ExpectationsStore", "name": "expectations_store"},
        {
            "class_name": "EvaluationParameterStore",
            "name": "evaluation_parameter_store",
        },
        {"class_name": "ValidationsStore", "name": "validation_result_store"},
    ]
    assert (
        basic_in_memory_data_context_just_stores.list_active_stores()
        == expected_store_list
    )

    store_name: str = "my_new_expectations_store"
    store_config: dict = {"class_name": "ExpectationsStore"}
    basic_in_memory_data_context_just_stores.add_store(
        store_name=store_name, store_config=store_config
    )
    assert (
        basic_in_memory_data_context_just_stores.list_active_stores()
        == expected_store_list
    )


def test_get_config_with_variables_substituted(
    basic_in_memory_data_context_just_stores,
):
    """
    Basic test for get_config_with_variables_substituted()
    A more thorough set of tests exist in test_data_context_config_variables.py with
    test_setting_config_variables_is_visible_immediately() testing whether the os.env values take
    precedence over config_file values, which they should.
    """

    context: BaseDataContext = basic_in_memory_data_context_just_stores
    assert isinstance(context.get_config(), DataContextConfig)

    # override the project config to use the $ escaped variable
    context._project_config["validations_store_name"] = "${replace_me}"
    try:
        # which we set from the environment
        os.environ["replace_me"] = "value_from_env_var"
        assert (
            context.get_config_with_variables_substituted().validations_store_name
            == "value_from_env_var"
        )
    finally:
        del os.environ["replace_me"]


@pytest.fixture
def prepare_validator_for_cloud_e2e() -> Callable[
    [CloudDataContext], Tuple[Validator, str]
]:
    def _closure(context: CloudDataContext) -> Tuple[Validator, str]:
        # Create a suite to be used in Validator instantiation
        suites = context.list_expectation_suites()
        expectation_suite_ge_cloud_id = suites[0].cloud_id

        # To ensure we don't accidentally impact parallel test runs in Azure, we randomly generate a suite name in this E2E test.
        # To limit the number of generated suites, we limit the randomization to 20 numbers.
        rand_suffix = random.randint(1, 20)
        suite_name = f"oss_e2e_test_suite_{rand_suffix}"

        # Start off each test run with a clean slate
        if expectation_suite_ge_cloud_id in context.list_expectation_suite_names():
            context.delete_expectation_suite(ge_cloud_id=expectation_suite_ge_cloud_id)

        suite = context.add_expectation_suite(
            suite_name,
            ge_cloud_id=expectation_suite_ge_cloud_id,
        )

        # Set up a number of Expectations and confirm proper assignment
        configs = [  # Content of configs do not matter as they are simply used to populate the suite (never run)
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist",
                kwargs={"column": "infinities"},
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist", kwargs={"column": "nulls"}
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_to_exist", kwargs={"column": "naturals"}
            ),
            ExpectationConfiguration(
                expectation_type="expect_column_values_to_be_unique",
                kwargs={"column": "naturals"},
            ),
        ]
        for config in configs:
            suite.add_expectation(expectation_configuration=config)

        suite.ge_cloud_id = expectation_suite_ge_cloud_id
        context.add_or_update_expectation_suite(
            expectation_suite=suite,
        )

        assert len(suite.expectations) == 4

        # Grab the first datasource/data connector/data asset bundle we can to use in Validation instantiation
        datasource_name = "Test Pandas Datasource"
        datasource = context.datasources[datasource_name]

        data_connector = tuple(datasource.data_connectors.values())[0]
        data_connector_name = data_connector.name

        data_asset_name = tuple(data_connector.assets.keys())[0]

        batch_request = RuntimeBatchRequest(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            runtime_parameters={
                "batch_data": pd.DataFrame({"x": range(10)}),
            },
            batch_identifiers={"col1": "123"},
        )

        # Create Validator and ensure that persistence is Cloud-backed
        validator = context.get_validator(
            batch_request=batch_request,
            expectation_suite_ge_cloud_id=expectation_suite_ge_cloud_id,
        )

        # Ensure that the Expectations set above propogate down successfully
        assert len(validator.expectation_suite.expectations) == 4

        return validator, expectation_suite_ge_cloud_id

    return _closure


@pytest.mark.xfail(
    reason="GX Cloud E2E tests are currently failing due to an id issue with ExpectationSuites; xfailing for purposes of the 0.15.20 release",
    run=True,
    strict=True,
)
@pytest.mark.e2e
@pytest.mark.cloud
@mock.patch("great_expectations.data_context.DataContext._save_project_config")
def test_get_validator_with_cloud_enabled_context_saves_expectation_suite_to_cloud_backend(
    mock_save_project_config: mock.MagicMock,
    prepare_validator_for_cloud_e2e: Callable[
        [CloudDataContext], Tuple[Validator, str]
    ],
) -> None:
    """
    What does this test do and why?

    Ensures that the Validator that is created through DataContext.get_validator() is
    Cloud-enabled if the DataContext used to instantiate the object is Cloud-enabled.
    Saving of ExpectationSuites using such a Validator should send payloads to the Cloud
    backend.
    """
    context = DataContext(cloud_mode=True)

    (
        validator,
        _,
    ) = prepare_validator_for_cloud_e2e(context)

    with mock.patch("requests.put", autospec=True) as mock_put:
        type(mock_put.return_value).status_code = mock.PropertyMock(return_value=200)
        validator.save_expectation_suite()

    assert mock_put.call_count == 1


@pytest.mark.xfail(
    reason="GX Cloud E2E tests are currently failing due to an id issue with ExpectationSuites; xfailing for purposes of the 0.15.20 release",
    run=True,
    strict=True,
)
@pytest.mark.e2e
@pytest.mark.cloud
@mock.patch("great_expectations.data_context.DataContext._save_project_config")
def test_validator_e2e_workflow_with_cloud_enabled_context(
    mock_save_project_config: mock.MagicMock,
    prepare_validator_for_cloud_e2e: Callable[
        [CloudDataContext], Tuple[Validator, str]
    ],
) -> None:
    """
    What does this test do and why?

    Ensures that the Validator that is created through DataContext.get_validator() is
    Cloud-enabled if the DataContext used to instantiate the object is Cloud-enabled.
    Saving of ExpectationSuites using such a Validator should send payloads to the Cloud
    backend.
    """
    context = DataContext(cloud_mode=True)

    (
        validator,
        expectation_suite_ge_cloud_id,
    ) = prepare_validator_for_cloud_e2e(context)

    assert len(validator.expectation_suite.expectations) == 4

    res = validator.expect_column_max_to_be_between("x", min_value=0, max_value=10)
    assert res.success

    assert len(validator.expectation_suite.expectations) == 5

    # Confirm that the suite is present before saving
    assert expectation_suite_ge_cloud_id in context.list_expectation_suite_names()
    suite_on_context = context.get_expectation_suite(
        ge_cloud_id=expectation_suite_ge_cloud_id
    )
    assert str(expectation_suite_ge_cloud_id) == str(suite_on_context.ge_cloud_id)

    validator.save_expectation_suite()
    assert len(validator.expectation_suite.expectations) == 5

    # Confirm that the suite is present after saving
    assert expectation_suite_ge_cloud_id in context.list_expectation_suite_names()
    suite_on_context = context.get_expectation_suite(
        ge_cloud_id=expectation_suite_ge_cloud_id
    )
    assert str(expectation_suite_ge_cloud_id) == str(suite_on_context.ge_cloud_id)
    assert len(suite_on_context.expectations) == 5



import pytest

from great_expectations.cli.datasource import sanitize_yaml_and_save_datasource
from great_expectations.util import is_library_loadable


def test_sanitize_yaml_and_save_datasource_raises_error_on_empty_yaml(
    empty_data_context,
):
    with pytest.raises(ValueError):
        sanitize_yaml_and_save_datasource(empty_data_context, "")


def test_sanitize_yaml_and_save_datasource_raises_error_on_non_string(
    empty_data_context,
):
    with pytest.raises(TypeError):
        for bad_input in [3, {"a", "b"}]:
            sanitize_yaml_and_save_datasource(empty_data_context, bad_input)


def test_sanitize_yaml_and_save_datasource_raises_error_on_missing_datasource_name(
    empty_data_context,
):
    yaml_snippet = """
class_name: SimpleSqlalchemyDatasource
introspection:
  whole_table:
    data_asset_name_suffix: __whole_table
connection_string: sqlite://"""

    with pytest.raises(ValueError):
        sanitize_yaml_and_save_datasource(empty_data_context, yaml_snippet)


def test_sanitize_yaml_and_save_datasource_works_without_credentials(
    sa,
    empty_data_context,
):
    context = empty_data_context
    yaml_snippet = """
name: my_datasource
class_name: SimpleSqlalchemyDatasource
introspection:
  whole_table:
    data_asset_name_suffix: __whole_table
connection_string: sqlite://"""

    assert len(context.list_datasources()) == 0
    sanitize_yaml_and_save_datasource(context, yaml_snippet)
    assert len(context.list_datasources()) == 1
    assert context.list_datasources() == [
        {
            "class_name": "SimpleSqlalchemyDatasource",
            "connection_string": "sqlite://",
            "introspection": {
                "whole_table": {"data_asset_name_suffix": "__whole_table"}
            },
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
        }
    ]
    obs = context.config_variables
    # remove the instance guid
    obs.pop("instance_id")
    assert obs == {}


@pytest.mark.skipif(
    not is_library_loadable(library_name="psycopg2"),
    reason="psycopg2 is not installed",
)
def test_sanitize_yaml_and_save_datasource_works_with_credentials(
    sa,
    empty_data_context,
):
    context = empty_data_context
    yaml_snippet = """
name: foo_datasource
class_name: SimpleSqlalchemyDatasource
credentials:
  host: localhost
  port: '5432'
  username: user
  password: pass
  database: postgres
  drivername: postgresql"""

    assert len(context.list_datasources()) == 0
    sanitize_yaml_and_save_datasource(context, yaml_snippet)
    assert len(context.list_datasources()) == 1
    assert context.list_datasources() == [
        {
            "class_name": "SimpleSqlalchemyDatasource",
            "credentials": {
                "database": "postgres",
                "drivername": "postgresql",
                "host": "localhost",
                "password": "***",
                "port": "5432",
                "username": "user",
            },
            "module_name": "great_expectations.datasource",
            "name": "foo_datasource",
        }
    ]
    obs = context.config_variables
    # remove the instance guid
    obs.pop("instance_id")
    assert obs == {
        "foo_datasource": {
            "database": "postgres",
            "drivername": "postgresql",
            "host": "localhost",
            "password": "pass",
            "port": "5432",
            "username": "user",
        }
    }


def test_sanitize_yaml_and_save_datasource_does_not_overwrite_duplicate_when_overwrite_existing_is_false(
    capsys,
    empty_data_context,
):
    context = empty_data_context
    pandas_yaml_snippet = """
name: my_datasource
class_name: Datasource
execution_engine:
  class_name: PandasExecutionEngine
"""

    assert len(context.list_datasources()) == 0
    sanitize_yaml_and_save_datasource(
        context, pandas_yaml_snippet, overwrite_existing=False
    )
    assert len(context.list_datasources()) == 1
    datasource_from_context = [
        {
            "class_name": "Datasource",
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
        }
    ]
    assert context.list_datasources() == datasource_from_context
    sanitize_yaml_and_save_datasource(
        context, pandas_yaml_snippet, overwrite_existing=False
    )
    assert len(context.list_datasources()) == 1

    stdout = capsys.readouterr().out.strip()
    assert (
        '**WARNING** A Datasource named "my_datasource" already exists in this Data Context. The Datasource has *not* been saved. Please use a different name or set overwrite_existing=True if you want to overwrite!'.strip()
        == stdout
    )

    # retest with a different type of datasource with the same name
    sql_yaml_snippet = """
    name: my_datasource
    class_name: SimpleSqlalchemyDatasource
    introspection:
      whole_table:
        data_asset_name_suffix: __whole_table
    connection_string: sqlite://"""

    sanitize_yaml_and_save_datasource(
        context, sql_yaml_snippet, overwrite_existing=False
    )
    assert len(context.list_datasources()) == 1

    stdout = capsys.readouterr().out.strip()
    assert (
        '**WARNING** A Datasource named "my_datasource" already exists in this Data Context. The Datasource has *not* been saved. Please use a different name or set overwrite_existing=True if you want to overwrite!'.strip()
        == stdout
    )


def test_sanitize_yaml_and_save_datasource_does_overwrite_duplicate_when_overwrite_existing_is_true(
    sa,
    capsys,
    empty_data_context,
):
    context = empty_data_context
    pandas_yaml_snippet = """
name: my_datasource
class_name: Datasource
execution_engine:
  class_name: PandasExecutionEngine
"""

    assert len(context.list_datasources()) == 0
    sanitize_yaml_and_save_datasource(
        context, pandas_yaml_snippet, overwrite_existing=True
    )
    assert len(context.list_datasources()) == 1
    pandas_datasource_from_context = [
        {
            "class_name": "Datasource",
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
            "execution_engine": {
                "class_name": "PandasExecutionEngine",
                "module_name": "great_expectations.execution_engine",
            },
        }
    ]
    assert context.list_datasources() == pandas_datasource_from_context
    sanitize_yaml_and_save_datasource(
        context, pandas_yaml_snippet, overwrite_existing=True
    )
    assert len(context.list_datasources()) == 1
    assert context.list_datasources() == pandas_datasource_from_context

    stdout = capsys.readouterr().out.strip()
    assert stdout == ""

    # retest with a different type of datasource with the same name
    sql_yaml_snippet = """
    name: my_datasource
    class_name: SimpleSqlalchemyDatasource
    introspection:
      whole_table:
        data_asset_name_suffix: __whole_table
    connection_string: sqlite://"""

    sanitize_yaml_and_save_datasource(
        context, sql_yaml_snippet, overwrite_existing=True
    )
    assert len(context.list_datasources()) == 1

    stdout = capsys.readouterr().out
    assert stdout == ""

    assert context.list_datasources() != pandas_datasource_from_context
    sql_datasource_from_context = [
        {
            "class_name": "SimpleSqlalchemyDatasource",
            "connection_string": "sqlite://",
            "introspection": {
                "whole_table": {"data_asset_name_suffix": "__whole_table"}
            },
            "module_name": "great_expectations.datasource",
            "name": "my_datasource",
        }
    ]
    assert context.list_datasources() == sql_datasource_from_context



"""These tests exercise ConfigurationBundle including Serialization."""
import pytest

from great_expectations.data_context import BaseDataContext
from great_expectations.data_context.migrator.configuration_bundle import (
    ConfigurationBundle,
    ConfigurationBundleJsonSerializer,
    ConfigurationBundleSchema,
)
from tests.data_context.migrator.conftest import StubBaseDataContext


@pytest.mark.cloud
@pytest.mark.unit
class TestConfigurationBundleCreate:
    def test_configuration_bundle_created(
        self,
        stub_base_data_context: StubBaseDataContext,
    ):
        """What does this test and why?

        Make sure the configuration bundle is created successfully from a data context.
        """

        context: BaseDataContext = stub_base_data_context

        config_bundle = ConfigurationBundle(context)

        assert config_bundle.is_usage_stats_enabled()
        assert config_bundle._data_context_variables is not None
        assert len(config_bundle.expectation_suites) == 1
        assert len(config_bundle.checkpoints) == 1
        assert len(config_bundle.profilers) == 1
        assert len(config_bundle.validation_results) == 1
        assert len(config_bundle.datasources) == 1

    def test_is_usage_statistics_key_set_if_key_not_present(
        self, stub_base_data_context_no_anonymous_usage_stats: StubBaseDataContext
    ):
        """What does this test and why?

        The ConfigurationBundle should handle a context that has not set the config for
         anonymous_usage_statistics.
        """
        context: BaseDataContext = stub_base_data_context_no_anonymous_usage_stats

        config_bundle = ConfigurationBundle(context)

        # If not supplied, an AnonymizedUsageStatisticsConfig is created in a
        # DataContextConfig and enabled by default.
        assert config_bundle.is_usage_stats_enabled()

    def test_configuration_bundle_created_usage_stats_disabled(
        self,
        stub_base_data_context_anonymous_usage_stats_present_but_disabled: StubBaseDataContext,
    ):
        """What does this test and why?

        Make sure the configuration bundle successfully parses the usage stats settings.
        """

        context: BaseDataContext = (
            stub_base_data_context_anonymous_usage_stats_present_but_disabled
        )

        config_bundle = ConfigurationBundle(context)

        assert not config_bundle.is_usage_stats_enabled()


@pytest.mark.cloud
@pytest.mark.unit
class TestConfigurationBundleSerialization:
    def test_configuration_bundle_serialization_all_fields(
        self,
        stub_base_data_context: StubBaseDataContext,
        stub_serialized_configuration_bundle: dict,
    ):
        """What does this test and why?

        Ensure configuration bundle is serialized correctly.
        """

        context: BaseDataContext = stub_base_data_context

        config_bundle = ConfigurationBundle(context)

        serializer = ConfigurationBundleJsonSerializer(
            schema=ConfigurationBundleSchema()
        )

        serialized_bundle: dict = serializer.serialize(config_bundle)

        expected_serialized_bundle = stub_serialized_configuration_bundle

        # Remove meta before comparing since it contains the GX version
        serialized_bundle["expectation_suites"][0].pop("meta", None)
        expected_serialized_bundle["expectation_suites"][0].pop("meta", None)

        assert serialized_bundle == expected_serialized_bundle

    def test_configuration_bundle_serialization_empty_fields(
        self,
        empty_serialized_configuration_bundle: dict,
    ):
        """What does this test and why?

        Ensure configuration bundle is serialized correctly.
        """

        context = StubBaseDataContext(
            checkpoint_names=tuple(),
            expectation_suite_names=tuple(),
            profiler_names=tuple(),
            validation_results_keys=tuple(),
            datasource_names=tuple(),
        )

        config_bundle = ConfigurationBundle(context)

        serializer = ConfigurationBundleJsonSerializer(
            schema=ConfigurationBundleSchema()
        )

        serialized_bundle: dict = serializer.serialize(config_bundle)

        assert serialized_bundle == empty_serialized_configuration_bundle

    def test_anonymous_usage_statistics_removed_during_serialization(
        self,
        stub_base_data_context: StubBaseDataContext,
    ):
        """What does this test and why?
        When serializing a ConfigurationBundle we need to remove the
        anonymous_usage_statistics key.
        """

        context: StubBaseDataContext = stub_base_data_context

        assert context.anonymous_usage_statistics is not None

        config_bundle = ConfigurationBundle(context)

        serializer = ConfigurationBundleJsonSerializer(
            schema=ConfigurationBundleSchema()
        )

        serialized_bundle: dict = serializer.serialize(config_bundle)

        assert (
            serialized_bundle["data_context_variables"].get(
                "anonymous_usage_statistics"
            )
            is None
        )



from typing import Optional

from great_expectations.core._docs_decorators import public_api
from great_expectations.core.expectation_configuration import (
    ExpectationConfiguration,  # noqa: TCH001
)
from great_expectations.expectations.expectation import (
    ColumnPairMapExpectation,
    InvalidExpectationConfigurationError,
)


class ExpectColumnPairValuesToBeInSet(ColumnPairMapExpectation):
    """Expect the paired values from columns A and B to belong to a set of valid pairs.

    expect_column_pair_values_to_be_in_set is a \
    [Column Pair Map Expectation](https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_column_pair_map_expectations).

    For example:
    ::
        >>> d = {'fruit': ['appple','apple','apple','banana','banana'],
                'color': ['red','green','yellow','yellow','red']}
        >>> my_df = pd.DataFrame(data=d)
        >>> my_df.expect_column_pair_values_to_be_in_set(
                'fruit',
                'color',
                [
                    ('apple','red'),
                    ('apple','green'),
                    ('apple','yellow'),
                    ('banana','yellow'),
                ]
        )
        {
            "success": false,
            "meta": {},
            "exception_info": {
                "raised_exception": false,
                "exception_traceback": null,
                "exception_message": null
            },
            "result": {
                "element_count": 5,
                "unexpected_count": 1,
                "unexpected_percent": 20.0,
                "partial_unexpected_list": [
                    [
                        "banana",
                        "red"
                    ]
                ],
                "missing_count": 0,
                "missing_percent": 0.0,
                "unexpected_percent_total": 20.0,
                "unexpected_percent_nonmissing": 20.0
            }
        }

    Args:
        column_A (str): The first column name
        column_B (str): The second column name
        value_pairs_set (list of tuples): All the valid pairs to be matched

    Keyword Args:
        ignore_row_if (str): "both_values_are_missing", "either_value_is_missing", "neither"

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        include_config (boolean): \
            If True, then include the expectation config as part of the result object.
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without modification. \
            For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, include_config, catch_exceptions, and meta.
    """

    # This dictionary contains metadata for display in the public gallery
    library_metadata = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "column pair map expectation",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    map_metric = "column_pair_values.in_set"
    success_keys = ("value_pairs_set", "ignore_row_if", "mostly")

    default_kwarg_values = {
        "mostly": 1.0,
        "ignore_row_if": "both_values_are_missing",
        "row_condition": None,
        "condition_parser": None,  # we expect this to be explicitly set whenever a row_condition is passed
        "result_format": "BASIC",
        "include_config": True,
        "catch_exceptions": False,
    }

    args_keys = (
        "column_A",
        "column_B",
        "value_pairs_set",
    )

    @public_api
    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """Validates the configuration of an Expectation.

        For `expect_column_pair_values_to_be_in_set` it is required that the `configuration.kwargs` contain `column_A`,
        `column_B`, and `value_pairs_set` keys.

        The configuration will also be validated using each of the `validate_configuration` methods in its Expectation
        superclass hierarchy.

        Args:
            configuration: An `ExpectationConfiguration` to validate. If no configuration is provided, it will be pulled
                from the configuration attribute of the Expectation instance.

        Raises:
            InvalidExpectationConfigurationError: The configuration does not contain the values required by the
                Expectation.
        """
        super().validate_configuration(configuration)
        configuration = configuration or self.configuration
        try:
            assert (
                "column_A" in configuration.kwargs
                and "column_B" in configuration.kwargs
            ), "both columns must be provided"
            assert (
                "value_pairs_set" in configuration.kwargs
            ), "must provide value_pairs_set"
        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))



import os.path
import sys
from pathlib import Path
from typing import Any, Optional

import pytest
from ruamel.yaml.error import YAMLStreamError

from great_expectations.core.yaml_handler import YAMLHandler


@pytest.fixture
def simple_yaml() -> str:
    simple_yaml: str = """
    name: test
    class_name: test_class
    module_name: test.test_class
    """
    return simple_yaml


@pytest.fixture
def simple_dict() -> dict:
    simple_dict: dict = {
        "name": "test",
        "class_name": "test_class",
        "module_name": "test.test_class",
    }
    return simple_dict


@pytest.fixture
def yaml_handler() -> YAMLHandler:
    return YAMLHandler()


@pytest.mark.unit
def test_load_correct_input(
    simple_yaml: str, simple_dict: dict, yaml_handler: YAMLHandler
) -> None:
    res: dict = yaml_handler.load(simple_yaml)

    assert res == simple_dict


@pytest.mark.unit
def test_load_incorrect_input(yaml_handler: YAMLHandler) -> None:
    with pytest.raises((TypeError, YAMLStreamError)):
        yaml_handler.load(12345)


@pytest.mark.integration
def test_file_output(tmp_path: Path, yaml_handler: YAMLHandler) -> None:
    simplest_yaml: str = "abc: 1"
    test_file: str = os.path.join(tmp_path, "out.yaml")
    out: Path = Path(test_file)

    data: dict = yaml_handler.load(simplest_yaml)
    yaml_handler.dump(data, out)

    # check the output
    with open(test_file) as f:
        line = f.readline().strip()
        data_from_file: dict = yaml_handler.load(line)

    assert data_from_file == data


@pytest.mark.unit
def test_dump_default_behavior_with_no_stream_specified(
    yaml_handler: YAMLHandler,
) -> None:
    # when we specify no stream, then StringIO is used by default
    simplest_dict: dict = dict(abc=1)
    dumped: Optional[str] = yaml_handler.dump(simplest_dict)
    assert dumped == "abc: 1\n"


@pytest.mark.unit
def test_dump_stdout_specified(capsys, yaml_handler: YAMLHandler) -> None:
    # ruamel documentation recommends that we specify the stream as stdout when we are using YAML to return a string.
    simplest_dict: dict = dict(abc=1)
    yaml_handler.dump(simplest_dict, stream=sys.stdout)
    captured: Any = capsys.readouterr()
    assert captured.out == "abc: 1\n"
